
import imp
import unittest
import os
import yaml
import numpy as np
import tensorflow as tf
from tensorflow.python.framework import graph_util
from neural_compressor.adaptor.tf_utils.util import disable_random
from tensorflow.python.ops import control_flow_ops


def build_fake_yaml():
    fake_yaml = '''
        model:
          name: fake_yaml
          framework: tensorflow
          inputs: input
          outputs: op_to_store
        device: cpu
        quantization:
          model_wise:
            weight:
                granularity: per_tensor
                scheme: sym
                dtype: int8
                algorithm: minmax
        evaluation:
          accuracy:
            metric:
              topk: 1
        tuning:
            strategy:
              name: mse
            accuracy_criterion:
              relative: 0.01
            exit_policy:
              performance_only: True
            workspace:
              path: saved
        '''
    y = yaml.load(fake_yaml, Loader=yaml.SafeLoader)
    with open('fake_yaml.yaml', "w", encoding="utf-8") as f:
        yaml.dump(y, f)
    f.close()


class TestSwitchOptimizer(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        build_fake_yaml()

    @classmethod
    def tearDownClass(self):
        os.remove('fake_yaml.yaml')

    @disable_random()
    def test_switch_optimizer(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        x_pad = tf.pad(x, paddings, "CONSTANT")
        y = tf.compat.v1.placeholder_with_default(True, [], name="place_true")
        
        conv_weights = tf.constant(np.random.random((3,3,16,16)).astype(np.float32), name='y')
        switch_false, switch_true = control_flow_ops.switch(conv_weights, y)
        conv = tf.nn.conv2d(x_pad, switch_true, strides=[1, 2, 2, 1], padding="VALID")
        normed = tf.compat.v1.layers.batch_normalization(conv)
        relu = tf.nn.relu(normed, name='op_to_store')
        out_name = relu.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

        from neural_compressor.adaptor.tf_utils.graph_rewriter.generic.switch_optimizer import SwitchOptimizer

        convert_graph = SwitchOptimizer(output_graph_def).do_transformation()
        found_switch = False
        for node in convert_graph.node:
          if node.op == 'Switch':
            found_switch = True
        self.assertEqual(found_switch, False)

    @disable_random()
    def test_switch_optimizer_invalid(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        x_pad = tf.pad(x, paddings, "CONSTANT")
        y = tf.compat.v1.placeholder_with_default(True, [], name="place_true")
        
        conv_weights = tf.constant(np.random.random((3,3,16,16)).astype(np.float32), name='y')
        switch_false, switch_true = control_flow_ops.switch(conv_weights, y)
        conv = tf.nn.conv2d(x_pad, switch_false, strides=[1, 2, 2, 1], padding="VALID")
        normed = tf.compat.v1.layers.batch_normalization(conv)
        relu = tf.nn.relu(normed, name='op_to_store')
        out_name = relu.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

        from neural_compressor.adaptor.tf_utils.graph_rewriter.generic.switch_optimizer import SwitchOptimizer

        convert_graph = SwitchOptimizer(output_graph_def).do_transformation()
        found_switch = False
        for node in convert_graph.node:
          if node.op == 'Switch':
            found_switch = True
        self.assertEqual(found_switch, True)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from onnxruntime.quantization.quant_utils import QuantizedValueType, \
        attribute_to_kwarg, ms_domain
from onnx import onnx_pb as onnx_proto
from neural_compressor.adaptor.ox_utils.util import QuantizedValue


class QLinearConcat(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node

        if all([item not in self.quantizer.quantized_value_map for item in node.input]):
            self.quantizer.new_nodes += [node]
            return

        if not all([item in self.quantizer.quantized_value_map for item in node.input]):
            super().quantize()
            return

        data_found, output_scale_name, output_zp_name, _, _ = \
                    self.quantizer._get_quantization_params(node.output[0])
        quantized_input_value = self.quantizer.quantized_value_map[node.input[0]]
        quantized_output_value = QuantizedValue(node.output[0], node.output[0] + "_quantized",
                                                output_scale_name, output_zp_name,
                                                quantized_input_value.value_type)
        self.quantizer.quantized_value_map[node.output[0]] = quantized_output_value

        (q_input_names, zero_point_names, scale_names, nodes) = self.quantizer.quantize_inputs(
                                                            node, [*range(0, len(node.input))])

        kwargs = {}
        for attribute in node.attribute:
            kwargs.update(attribute_to_kwarg(attribute))
        kwargs["domain"] = ms_domain
        qnode_name = node.name + "_quant" if node.name != "" else ""

        qlconcat_inputs = [output_scale_name, output_zp_name]
        for i in range(0, len(q_input_names)):
            qlconcat_inputs.extend([q_input_names[i], scale_names[i], zero_point_names[i]])
        qlconcat_node = onnx.helper.make_node("QLinearConcat", 
                                              qlconcat_inputs, 
                                              [node.output[0] + "_quantized"], 
                                              qnode_name, 
                                              **kwargs)        

        self.quantizer.new_nodes += nodes
        self.quantizer.new_nodes += [qlconcat_node]



#
#  -*- coding: utf-8 -*-
#
import unittest
import os
import tensorflow as tf
from neural_compressor.adaptor.tf_utils.util import get_estimator_graph

class TestEstimatorGraphConvert(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.dst_path = '/tmp/.neural_compressor/train.csv'
        self.titanic_file = tf.keras.utils.get_file(self.dst_path, \
            "https://storage.googleapis.com/tf-datasets/titanic/train.csv")

    def test_get_estimator_graph(self):
        def train_input_fn():
          titanic = tf.data.experimental.make_csv_dataset(
              self.titanic_file, batch_size=32,
              label_name="survived")
          titanic_batches = (
              titanic.cache().repeat().shuffle(500)
              .prefetch(tf.data.experimental.AUTOTUNE))
          return titanic_batches
        age = tf.feature_column.numeric_column('age')
        cls = tf.feature_column.categorical_column_with_vocabulary_list('class', \
            ['First', 'Second', 'Third']) 
        embark = tf.feature_column.categorical_column_with_hash_bucket('embark_town', 32)
        import tempfile
        model_dir = tempfile.mkdtemp()
        model = tf.estimator.LinearClassifier(
            model_dir=model_dir,
            feature_columns=[embark, cls, age],
            n_classes=2
        )
        model = model.train(input_fn=train_input_fn, steps=100)
        result = model.evaluate(train_input_fn, steps=10)

        graph = get_estimator_graph(model, train_input_fn)

        self.assertTrue(isinstance(graph, tf.Graph)) 
        graph_def = graph.as_graph_def()
        self.assertGreater(len(graph_def.node), 1)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration type parser."""
import json
from collections.abc import Iterable
from typing import Any, Dict, List, Type, Union

from neural_compressor.ux.utils.exceptions import ClientErrorException
from neural_compressor.ux.utils.hw_info import HWInfo
from neural_compressor.ux.utils.utils import parse_bool_value


class ConfigurationParser:
    """Configuration type parser class."""

    def __init__(self) -> None:
        """Initialize configuration type parser."""
        self.transform_types: Dict[str, List[str]] = {
            "str": ["interpolation", "dtype", "label_file", "vocab_file"],
            "int": [
                "x",
                "y",
                "height",
                "width",
                "offset_height",
                "offset_width",
                "target_height",
                "target_width",
                "dim",
                "resize_side",
                "label_shift",
            ],
            "float": ["scale", "central_fraction"],
            "list<float>": ["mean", "std", "mean_value", "std_value", "ratio"],
            "list<int>": ["perm", "size"],
            "bool": ["random_crop", "random_flip_left_right"],
        }
        self.dataloader_types: Dict[str, List[str]] = {
            "str": [
                "root",
                "filenames",
                "compression_type",
                "data_path",
                "image_list",
                "img_dir",
                "anno_dir",
                "content_folder",
                "style_folder",
                "image_format",
                "dtype",
                "label_file",
            ],
            "int": ["buffer_size", "num_parallel_reads", "num_cores"],
            "list<int>": ["resize_shape"],
            "float": ["crop_ratio"],
            "list<float>": ["low", "high"],
            "list<list<int>>": ["shape"],
            "bool": ["train", "label"],
        }
        self.types_definitions: Dict[str, Union[Type, List[Any]]] = {
            "str": str,
            "int": int,
            "list<int>": [int],
            "list<list<int>>": [[int]],
            "float": float,
            "list<float>": [float],
            "bool": bool,
        }

    def parse(self, data: dict) -> dict:
        """Parse configuration."""
        data = set_defaults(data)

        transforms_data = data.get("transform", None)
        if transforms_data is not None:
            data.update({"transform": self.parse_transforms(transforms_data)})

        quantization_dataloader = data.get("quantization", {}).get("dataloader", None)
        if quantization_dataloader and isinstance(quantization_dataloader, dict):
            data["quantization"].update(
                {"dataloader": self.parse_dataloader(quantization_dataloader)},
            )

        evaluation_dataloader = data.get("evaluation", {}).get("dataloader", None)
        if evaluation_dataloader and isinstance(evaluation_dataloader, dict):
            data["evaluation"].update(
                {"dataloader": self.parse_dataloader(evaluation_dataloader)},
            )

        num_cores = HWInfo().cores
        cores_per_instance = int(data.get("evaluation", {}).get("cores_per_instance", 4))

        if cores_per_instance < 1:
            raise ClientErrorException(
                "At least one core per instance must be used.",
            )
        if cores_per_instance > num_cores:
            raise ClientErrorException(
                f"Requested {cores_per_instance} cores per instance, "
                f"while only {num_cores} available.",
            )

        max_number_of_instances = num_cores // cores_per_instance
        instances = int(data.get("evaluation", {}).get("instances", max_number_of_instances))

        if instances < 1:
            raise ClientErrorException("At least one instance must be used.")

        if instances > max_number_of_instances:
            raise ClientErrorException(
                f"Attempted to use {instances} instances, "
                f"while only {max_number_of_instances} allowed.",
            )

        if "evaluation" in data:
            data["evaluation"].update(
                {
                    "cores_per_instance": cores_per_instance,
                    "num_of_instance": instances,
                    "batch_size": int(data.get("evaluation", {}).get("batch_size", 1)),
                },
            )

        data["tuning"] = parse_bool_value(data["tuning"])

        return data

    def parse_transforms(self, transforms_data: List[dict]) -> List[dict]:
        """Parse transforms list."""
        for transform in transforms_data:
            params_to_parse = transform.get("params", None)
            if isinstance(params_to_parse, dict):
                for param_name, value in params_to_parse.items():
                    param_type: Union[Type, List[Type]] = self.get_param_type(
                        "transform",
                        param_name,
                    )
                    if transform.get("name") == "RandomResizedCrop" and param_name == "scale":
                        param_type = [float]
                    transform["params"].update(
                        {param_name: self.parse_value(value, param_type)},
                    )
        return transforms_data

    def parse_dataloader(self, dataloader_data: dict) -> dict:
        """Parse dataloader dict."""
        dataloader_params = dataloader_data.get("params", None)
        if isinstance(dataloader_params, dict):
            for param_name, value in dataloader_params.items():
                param_type: Union[Type, List[Type]] = self.get_param_type(
                    "dataloader",
                    param_name,
                )
                dataloader_data["params"].update(
                    {param_name: self.parse_value(value, param_type)},
                )
        return dataloader_data

    def get_param_type(
        self,
        param_group: str,
        param_name: str,
    ) -> Union[Type, List[Type]]:
        """Get parameter type."""
        params_definitions = {}
        if param_group == "transform":
            params_definitions = self.transform_types
        elif param_group == "dataloader":
            params_definitions = self.dataloader_types
        for param_type, param_names in params_definitions.items():
            if param_name in param_names:
                found_type = self.types_definitions.get(param_type, None)
                if found_type is not None:
                    return found_type
        raise Exception(
            f"Could not found type for {param_group} {param_name} parameter.",
        )

    @staticmethod
    def parse_value(value: Any, required_type: Union[Type, List[Type]]) -> Any:
        """Parse value to required type."""
        try:
            if required_type == bool:
                return parse_bool_value(value)
            if callable(required_type):
                return required_type(value)
            elif isinstance(required_type, list):
                return parse_list_value(value, required_type[0])
        except ValueError as err:
            raise ClientErrorException(f"Cannot cast {value}. {str(err)}")
        return value


def parse_list_value(value: Any, required_type: Union[Type, List[Type]]) -> List[Any]:
    """Parse value to list."""
    if isinstance(required_type, list):
        return parse_multidim_list(value, required_type[0])
    if isinstance(value, str):
        return [required_type(element.strip("")) for element in value.strip("[]").split(",")]
    elif isinstance(value, Iterable):
        return [required_type(item) for item in value]
    elif callable(required_type):
        return [required_type(value)]
    else:
        return [value]


def parse_multidim_list(value: Any, required_type: Type) -> List[Union[Any, List[Any]]]:
    """Parse multi dimensional list."""
    if isinstance(value, str):
        value = normalize_string_list(value)
        parsed_list = json.loads(value)
    else:
        parsed_list = value

    if callable(required_type):
        for top_idx, top_element in enumerate(parsed_list):
            if isinstance(top_element, list):
                for idx, element in enumerate(top_element):
                    parsed_list[top_idx][idx] = required_type(element)
            else:
                parsed_list[top_idx] = required_type(top_element)
    return parsed_list


def normalize_string_list(string_list: str) -> str:
    """Add wrap string list into brackets if missing."""
    if not isinstance(string_list, str):
        return string_list
    if not string_list.startswith("["):
        string_list = "[" + string_list
    if not string_list.endswith("]"):
        string_list += "]"
    return string_list


def set_defaults(data: dict) -> dict:
    """Set default values for data if missing."""
    # Set tuning as default
    if "tuning" not in data:
        data.update({"tuning": True})

    # Set int8 as default requested precision
    if "precision" not in data:
        data.update({"precision": "int8"})

    if not data["tuning"]:
        data["dataset_path"] = "no_dataset_location"

    return data



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Environment manager class."""
import os
import sys

from neural_compressor.ux.utils.templates.workdir import Workdir
from neural_compressor.ux.utils.workload.workload import WorkloadMigrator
from neural_compressor.ux.utils.workload.workloads_list import WorkloadsListMigrator


class Environment:
    """Environment manager class."""

    @staticmethod
    def ensure_workdir_exists_and_writeable() -> None:
        """Ensure that configured directory exists and can be used."""
        from neural_compressor.ux.utils.logger import log
        from neural_compressor.ux.web.configuration import Configuration

        configuration = Configuration()
        workdir = configuration.workdir
        error_message_tail = "Please ensure it is a directory that can be written to.\nExiting.\n"
        try:
            os.makedirs(workdir, exist_ok=True)
        except Exception as e:
            print(f"Unable to create workdir at {workdir}: {e}.\n{error_message_tail}")
            log.error(e)
            sys.exit(1)
        if not os.access(workdir, os.W_OK):
            print(f"Unable to create files at {workdir}.\n{error_message_tail}")
            sys.exit(2)

    @staticmethod
    def migrate_workloads_list() -> None:
        """Migrate workloads list to latest format version."""
        workload_list_migrator = WorkloadsListMigrator()
        if workload_list_migrator.require_migration:
            workload_list_migrator.migrate()
            workload_list_migrator.dump()

    @staticmethod
    def migrate_workloads() -> None:
        """Migrate workloads to latest format version."""
        workload_list_migrator = WorkloadsListMigrator()
        if not os.path.isfile(workload_list_migrator.workloads_json):
            return
        workload_list_migrator.load_workloads_data()
        updated_workloads = {}
        for workload_id, workload_data in workload_list_migrator.workloads_data.get(
            "workloads",
            {},
        ).items():
            try:
                workload_path = workload_data.get("workload_path", None)
                if workload_path is None:
                    continue
                workload_json_path = os.path.join(workload_path, "workload.json")
                workload_migrator = WorkloadMigrator(
                    workload_json_path=workload_json_path,
                )
                workload_migrator.migrate()
                workload_migrator.dump()
                updated_workloads[workload_id] = workload_data
            except Exception:
                pass
        workload_list_migrator.workloads_data["workloads"] = updated_workloads
        workload_list_migrator.dump()

    @staticmethod
    def clean_workloads_wip_status() -> None:
        """Clean WIP status for workloads in workloads_list.json."""
        workdir = Workdir()
        workdir.clean_status(status_to_clean="wip")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import numpy as np
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_util
from neural_compressor.utils.utility import dump_elapsed_time
from neural_compressor.adaptor.tf_utils.graph_rewriter.graph_util import GraphAnalyzer
from ..graph_base import GraphRewriterBase
from ..graph_util import GraphRewriterHelper as Helper


class GenerateITEXModel(GraphRewriterBase):
    """ Insert Q/DQ pairs before quantizable ops.

    Args: model: input model.
          data: sampling data.

    Return: converted model
    """

    def __init__(self, model, calibration_data):
        super().__init__(model)
        self.data = calibration_data

    @dump_elapsed_time("Pass GenerateITEXGraph")
    def do_transformation(self):
        itex_min_max_values = {}
        for i in self.data:
            if i.find('_requant') == -1:
                key, value = i.rsplit(':')[0], i.rsplit(':')[1]
                key = key.split('_eightbit_')[0][1:] + key[-5:]
                if key not in itex_min_max_values:
                    itex_min_max_values[key] = [float(value[1:-1])]
                else:
                    itex_min_max_values[key].append(float(value[1:-1]))
        quantizable_op_names = []
        for i in itex_min_max_values:
            if i.split('__')[0] not in quantizable_op_names:
                quantizable_op_names.append(i.split('__')[0])

        g = GraphAnalyzer()
        g.graph = copy.deepcopy(self.model.graph_def)
        graph_info = g.parse_graph()
        for op_name in quantizable_op_names:
            min_node = Helper.create_constant_node(
                op_name + '_min', np.min(itex_min_max_values[op_name+'__min']), dtypes.float32)
            max_node = Helper.create_constant_node(
                op_name + '_max', np.max(itex_min_max_values[op_name+'__max']), dtypes.float32)
            quantizable_node_input = graph_info[op_name].node.input[0]
            quant_v2_node = Helper.create_node(
                "QuantizeV2", op_name + '_quantize',
                [quantizable_node_input, min_node.name, max_node.name])
            is_asymmetric = bool(graph_info[op_name].node.op == 'MatMul')
            quant_deq_dt = dtypes.quint8 if is_asymmetric else dtypes.qint8
            Helper.set_attr_dtype(quant_v2_node, "T", quant_deq_dt)
            if not is_asymmetric:
                Helper.set_attr_string(quant_v2_node, "round_mode", b"HALF_TO_EVEN")
            Helper.set_attr_bool(quant_v2_node, "narrow_range", False if is_asymmetric else True)

            Helper.set_attr_string(
                quant_v2_node, "mode", b"MIN_FIRST" if is_asymmetric else b"SCALED")

            dequantize_node = Helper.create_node("Dequantize", op_name + '_dequantize',
                                                 [quant_v2_node.name,
                                                 quant_v2_node.name + ':1',
                                                 quant_v2_node.name + ':2'])
            Helper.set_attr_dtype(dequantize_node, "T", quant_deq_dt)
            Helper.set_attr_string(
                dequantize_node, "mode", b"MIN_FIRST" if is_asymmetric else b"SCALED")
            g.add_node(quant_v2_node,
                        graph_info[op_name].node.input[0],
                        [dequantize_node.name])
            g.add_node(dequantize_node, quant_v2_node.name, [op_name])
            g.add_node(min_node, None, [quant_v2_node.name])
            g.add_node(max_node, None, [quant_v2_node.name])
            graph_info[op_name].node.input[0] = dequantize_node.name

        g_weight = GraphAnalyzer()
        g_weight.graph = g.dump_graph()
        graph_info = g_weight.parse_graph()
        target_nodes = g_weight.query_fusion_pattern_nodes(
            [["Conv2D", "MatMul"], ["BiasAdd"], ('Relu',)])
        for i in target_nodes:
            computational_node_name = i[0]
            computational_node = graph_info[computational_node_name].node
            weight_name = computational_node.input[1]
            weight_node = graph_info[weight_name].node
            weight_tensor = tensor_util.MakeNdarray(weight_node.attr['value'].tensor)
            min_value = np.min(weight_tensor)
            max_value = np.max(weight_tensor)
            min_const_node = Helper.create_constant_node(
                weight_name + '_min', min_value, dtypes.float32)
            max_const_node = Helper.create_constant_node(
                weight_name + '_max', max_value, dtypes.float32)
            quant_node = Helper.create_node(
                "QuantizeV2", weight_name + '_quant',
                [weight_name, weight_name + '_min', weight_name + '_max'])
            dequant_node=Helper.create_node(
                "Dequantize", weight_name + '_dequant',
                [quant_node.name, quant_node.name + ':1', quant_node.name + ':2'])
            Helper.set_attr_dtype(quant_node, "T", dtypes.qint8)
            Helper.set_attr_string(quant_node, "mode", b"SCALED")
            Helper.set_attr_string(quant_node, "round_mode", b"HALF_TO_EVEN")

            Helper.set_attr_dtype(dequant_node, "T", dtypes.qint8)
            Helper.set_attr_string(dequant_node, "mode", b"SCALED")

            g_weight.add_node(quant_node, weight_name, [])
            g_weight.add_node(min_const_node, None, [quant_node.name])
            g_weight.add_node(max_const_node, None, [quant_node.name])
            g_weight.add_node(dequant_node, quant_node.name, [computational_node_name])
            computational_node.input[1] = dequant_node.name

        return g_weight.dump_graph()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration type parser test."""


import unittest
from typing import Any, Dict, List, Optional, Union
from unittest.mock import MagicMock, patch

from neural_compressor.ux.components.configuration_wizard.configuration_parser import (
    ConfigurationParser,
)
from neural_compressor.ux.utils.exceptions import ClientErrorException


@patch("sys.argv", ["neural_compressor_bench.py", "-p5000"])
class TestParser(unittest.TestCase):
    """Main test class for parser."""

    def __init__(self, *args: str, **kwargs: str) -> None:
        """Parser tests constructor."""
        super().__init__(*args, **kwargs)
        self.parser = ConfigurationParser()

    def _assert_parses_to_expected_float(self, value: Any, expected: float) -> None:
        self._assert_parses_to_expected_type(value, expected, float)

    def _assert_parses_float_failure(self, value: Any) -> None:
        self._assert_parses_failure(value, float)

    def _assert_parses_to_expected_int(self, value: Any, expected: int) -> None:
        self._assert_parses_to_expected_type(value, expected, int)

    def _assert_parses_int_failure(self, value: Any) -> None:
        self._assert_parses_failure(value, int)

    def _assert_parses_to_expected_bool(self, value: Any, expected: bool) -> None:
        self._assert_parses_to_expected_type(value, expected, bool)

    def _assert_parses_bool_failure(self, value: Any) -> None:
        self._assert_parses_failure(value, bool)

    def _assert_parses_to_expected_list(
        self,
        value: Any,
        element_type: type,
        expected: list,
    ) -> None:
        self._assert_parses_to_expected_list_type(value, expected, [element_type])

    def _assert_parses_to_expected_type(
        self,
        value: Any,
        expected: Any,
        required_type: Union[type, List[type]],
    ) -> None:
        parsed_value = self.parser.parse_value(value, required_type)
        self.assertEqual(parsed_value, expected)
        self.assertIs(type(parsed_value), required_type)

    def _assert_parses_failure(
        self,
        value: Any,
        required_type: Union[type, List[type]],
    ) -> None:
        """Test failed parsing."""
        with self.assertRaises(ClientErrorException):
            self.parser.parse_value(value, required_type)

    def _assert_parses_to_expected_list_type(
        self,
        value: Any,
        expected: Any,
        required_type: List[type],
    ) -> None:
        parsed_value = self.parser.parse_value(value, required_type)
        self.assertEqual(parsed_value, expected)
        self.assertIs(type(parsed_value), list)
        for element in parsed_value:
            self.assertIs(type(element), required_type[0])


class TestValueParser(TestParser):
    """Value parser tests."""

    def __init__(self, *args: str, **kwargs: str) -> None:
        """Parser tests constructor."""
        super().__init__(*args, **kwargs)

    def test_float_to_float(self) -> None:
        """Test float value parsing."""
        self._assert_parses_to_expected_float(0.1, 0.1)

    def test_int_to_float(self) -> None:
        """Test int value parsing."""
        self._assert_parses_to_expected_float(1, 1.0)

    def test_string_to_float(self) -> None:
        """Test parsing string to float."""
        self._assert_parses_to_expected_float("0.1", 0.1)

    def test_string_with_whitespace_to_float(self) -> None:
        """Test parsing string with whitespace to float."""
        self._assert_parses_to_expected_float(" \t 0.1 ", 0.1)

    def test_float_parsing_failure(self) -> None:
        """Test parsing failure."""
        self._assert_parses_float_failure("abc")

    def test_float_to_int(self) -> None:
        """Test float value parsing to int."""
        self._assert_parses_to_expected_int(11.1, 11)

    def test_int_to_int(self) -> None:
        """Test int value parsing."""
        self._assert_parses_to_expected_int(11, 11)

    def test_string_to_int(self) -> None:
        """Test parsing string to int."""
        self._assert_parses_to_expected_int("11", 11)

    def test_string_with_whitespace_to_int(self) -> None:
        """Test parsing string with whitespace to int."""
        self._assert_parses_to_expected_int(" \t \t 11 \t \t ", 11)

    def test_int_parsing_failure(self) -> None:
        """Test parsing failure."""
        self._assert_parses_int_failure("abc")

    def test_float_to_list(self) -> None:
        """Test float value parsing to float list."""
        self._assert_parses_to_expected_list(11.1, float, [11.1])

    def test_int_to_float_list(self) -> None:
        """Test int value parsing to float list."""
        self._assert_parses_to_expected_list(11, float, [11])

    def test_int_to_list(self) -> None:
        """Test int value parsing to int list."""
        self._assert_parses_to_expected_list(11, int, [11])

    def test_string_to_list(self) -> None:
        """Test parsing string to int list."""
        self._assert_parses_to_expected_list("11", int, [11])

    def test_string_with_brackets_to_list(self) -> None:
        """Test parsing string to list."""
        self._assert_parses_to_expected_list("[3, 14, 15, 9]", int, [3, 14, 15, 9])

    def test_string_to_bool_false_lowercase(self) -> None:
        """Test parsing 'false' to boolean."""
        self._assert_parses_to_expected_bool("false", False)

    def test_string_to_bool_false_capitalized(self) -> None:
        """Test parsing 'False' to boolean."""
        self._assert_parses_to_expected_bool("False", False)

    def test_string_to_bool_false_mixed_case(self) -> None:
        """Test parsing 'FaLsE' to boolean."""
        self._assert_parses_to_expected_bool("FaLsE", False)

    def test_string_to_bool_false_mixed_case_with_whitespace(self) -> None:
        """Test parsing 'FaLsE' to boolean."""
        self._assert_parses_to_expected_bool("   FaLsE \t", False)

    def test_string_to_bool_true_lowercase(self) -> None:
        """Test parsing 'true' to boolean."""
        self._assert_parses_to_expected_bool("true", True)

    def test_string_to_bool_true_capitalized(self) -> None:
        """Test parsing 'True' to boolean."""
        self._assert_parses_to_expected_bool("True", True)

    def test_string_to_bool_true_mixed_case(self) -> None:
        """Test parsing 'TrUe' to boolean."""
        self._assert_parses_to_expected_bool("TrUe", True)

    def test_string_to_bool_true_mixed_case_with_whitespace(self) -> None:
        """Test parsing 'TrUe' to boolean."""
        self._assert_parses_to_expected_bool("\t TrUe ", True)

    def test_bool_parsing_failure(self) -> None:
        """Test parsing failure."""
        self._assert_parses_bool_failure("\nsi")


class TestTransformParser(TestParser):
    """Transform parser tests."""

    def __init__(self, *args: str, **kwargs: str) -> None:
        """Parser tests constructor."""
        super().__init__(*args, **kwargs)

    def test_parses_AlignImageChannel(self) -> None:
        """Test parsing of AlignImageChannel parameters."""
        self._assert_parses_single_transform_params(
            "AlignImageChannel",
            {"dim": "123"},
            {"dim": 123},
        )

    def test_parses_default_BilinearImagenet(self) -> None:
        """Test parsing of default BilinearImagenet parameters."""
        self._assert_parses_single_transform_params(
            "BilinearImagenet",
            {
                "height": 224,
                "width": 224,
                "central_fraction": 0.875,
                "mean_value": "[0.0, 0.0, 0.0]",
                "scale": 1,
            },
            {
                "height": 224,
                "width": 224,
                "central_fraction": 0.875,
                "mean_value": [0.0, 0.0, 0.0],
                "scale": 1,
            },
        )

    def test_parses_BilinearImagenet(self) -> None:
        """Test parsing of BilinearImagenet parameters."""
        self._assert_parses_single_transform_params(
            "BilinearImagenet",
            {
                "height": "1",
                "width": "2",
                "central_fraction": "3.14",
                "mean_value": "[4.5]",
                "scale": "6.7",
            },
            {
                "height": 1,
                "width": 2,
                "central_fraction": 3.14,
                "mean_value": [4.5],
                "scale": 6.7,
            },
        )

    def test_parses_Cast(self) -> None:
        """Test parsing of Cast parameters."""
        self._assert_parses_single_transform_params(
            "Cast",
            {"dtype": "float32"},
            {"dtype": "float32"},
        )

    def test_parses_empty_data(self) -> None:
        """Test parsing data without transforms."""
        parsed = self.parser.parse({})

        self.assertIsNone(parsed.get("transform", None))

    def _build_input_with_transform(self, name: str, params: Optional[dict]) -> dict:
        """Build fake data with single Transform."""
        return {
            "domain": "image_recognition",
            "framework": "tensorflow",
            "id": "f435c2a528c38176d1bb829e5cff3690",
            "model_path": "/foo/bar/baz.pb",
            "inputs": ["input_tensor"],
            "outputs": ["softmax_tensor"],
            "transform": [{"name": name, "params": params}],
            "quantization": {
                "dataset_path": "/foo/bar/baz.pb",
                "dataloader": {"name": "ImageRecord", "params": {"root": ""}},
                "accuracy_goal": 0.01,
                "sampling_size": 100,
                "strategy": "basic",
                "approach": "post_training_static_quant",
                "objective": "",
                "timeout": 0,
                "max_trials": 100,
                "random_seed": 9527,
            },
            "evaluation": {
                "dataset_path": "/foo/bat/baz",
                "dataloader": {"name": "ImageRecord", "params": {"root": ""}},
                "metric": "topk",
                "metric_param": 1,
                "batch_size": 1,
                "cores_per_instance": "1",
                "instances": "1",
                "inter_nr_of_threads": "",
                "intra_nr_of_threads": "",
                "iterations": -1,
                "warmup": 10,
                "kmp_blocktime": 1,
            },
            "workspace_path": "/foo",
            "tuning": True,
        }

    def _build_input_with_multiple_transforms(
        self,
        transform_list: List[Dict[str, Union[str, Optional[dict]]]],
    ) -> dict:
        """Build fake data with multiple transforms."""
        return {
            "domain": "image_recognition",
            "framework": "tensorflow",
            "id": "f435c2a528c38176d1bb829e5cff3690",
            "model_path": "/foo/bar/baz.pb",
            "inputs": ["input_tensor"],
            "outputs": ["softmax_tensor"],
            "transform": transform_list,
            "quantization": {
                "dataset_path": "/foo/bar/baz.pb",
                "dataloader": {"name": "ImageRecord", "params": {"root": ""}},
                "accuracy_goal": 0.01,
                "sampling_size": 100,
                "strategy": "basic",
                "approach": "post_training_static_quant",
                "objective": "",
                "timeout": 0,
                "max_trials": 100,
                "random_seed": 9527,
            },
            "evaluation": {
                "dataset_path": "/foo/bat/baz",
                "dataloader": {"name": "ImageRecord", "params": {"root": ""}},
                "metric": "topk",
                "metric_param": 1,
                "batch_size": 1,
                "cores_per_instance": "1",
                "instances": "1",
                "inter_nr_of_threads": "",
                "intra_nr_of_threads": "",
                "iterations": -1,
                "warmup": 10,
                "kmp_blocktime": 1,
            },
            "workspace_path": "/foo",
            "tuning": True,
        }

    def _assert_parses_single_transform_params(
        self,
        name: str,
        input_params: Optional[dict],
        expected_params: Optional[dict],
    ) -> None:
        parsed = self.parser.parse(self._build_input_with_transform(name, input_params))
        transforms = parsed.get("transform", None)
        self.assertIsNotNone(transforms)
        matching_transforms = list(
            filter(
                lambda e: name == e.get("name", ""),
                transforms,
            ),
        )
        self.assertEqual(1, len(matching_transforms))
        parsed_params = matching_transforms[0].get("params", {})
        self.assertEqual(parsed_params, expected_params)

    def _assert_parses_multiple_transforms(
        self,
        input_list: List[Dict[str, Union[str, Optional[dict]]]],
        expected_list: List[Dict[str, Union[str, Optional[dict]]]],
    ) -> None:
        parsed = self.parser.parse(
            self._build_input_with_multiple_transforms(input_list),
        )
        transforms = parsed.get("transform", None)
        self.assertIsNotNone(transforms)
        self.assertCountEqual(transforms, expected_list)
        self.assertListEqual(transforms, expected_list)


class TestDataloaderParser(TestParser):
    """Dataloader parser tests."""

    def __init__(self, *args: str, **kwargs: str) -> None:
        """Parser tests constructor."""
        super().__init__(*args, **kwargs)

    def test_parses_dummy_dataloader(self) -> None:
        """Test parsing dummy dataloader."""
        self._assert_parses_dataloader_params(
            "dummy",
            {
                "dtype": "float32",
                "high": 127,
                "label": "False",
                "low": -128,
                "shape": "[224,224]",
            },
            {
                "dtype": "float32",
                "high": [127],
                "label": False,
                "low": [-128],
                "shape": [224, 224],
            },
        )

    def test_parses_dataloader_shape_without_brackets(self) -> None:
        """Test parsing dummy dataloader."""
        self._assert_parses_dataloader_params(
            "dummy",
            {
                "dtype": "float32",
                "high": 127,
                "label": "False",
                "low": -128,
                "shape": "224,224",
            },
            {
                "dtype": "float32",
                "high": [127],
                "label": False,
                "low": [-128],
                "shape": [224, 224],
            },
        )

    def test_parses_multidim_dummy_dataloader(self) -> None:
        """Test parsing multidim dummy dataloader."""
        self._assert_parses_dataloader_params(
            "dummy",
            {
                "shape": "[[128, 3, 224, 224], [128, 1, 1, 1]]",
            },
            {
                "shape": [[128, 3, 224, 224], [128, 1, 1, 1]],
            },
        )

    def test_parses_evaluation_performance_dataloader_batch_size(self) -> None:
        """Test parsing evaluation.performance.batch_size."""
        parsed = self.parser.parse(
            self._build_input_with_dataloader("foo", {}),
        )

        self.assertEqual(234, parsed.get("evaluation", {}).get("batch_size"))

    def _build_input_with_dataloader(self, name: str, params: Optional[dict]) -> dict:
        """Build fake data with dataloader."""
        return {
            "domain": "image_recognition",
            "framework": "tensorflow",
            "id": "f435c2a528c38176d1bb829e5cff3690",
            "model_path": "/foo/bar/baz.pb",
            "inputs": ["input_tensor"],
            "outputs": ["softmax_tensor"],
            "transform": [],
            "quantization": {
                "dataset_path": "/foo/bar/baz.pb",
                "dataloader": {"name": name, "params": params},
                "accuracy_goal": 0.01,
                "sampling_size": 100,
                "strategy": "basic",
                "approach": "post_training_static_quant",
                "objective": "",
                "timeout": 0,
                "max_trials": 100,
                "random_seed": 9527,
            },
            "evaluation": {
                "dataset_path": "/foo/bat/baz",
                "dataloader": {"name": name, "params": params},
                "metric": "topk",
                "metric_param": 1,
                "batch_size": "234",
                "cores_per_instance": "1",
                "instances": "1",
                "inter_nr_of_threads": "",
                "intra_nr_of_threads": "",
                "iterations": -1,
                "warmup": 10,
                "kmp_blocktime": 1,
            },
            "workspace_path": "/foo",
            "tuning": True,
        }

    def _assert_parses_dataloader_params(
        self,
        name: str,
        input_params: Optional[dict],
        expected_params: Optional[dict],
    ) -> None:
        parsed = self.parser.parse(
            self._build_input_with_dataloader(name, input_params),
        )
        q_dataloader_params = (
            parsed.get("quantization", {}).get("dataloader", {}).get("params", None)
        )
        eval_dataloader_params = (
            parsed.get("evaluation", {}).get("dataloader", None).get("params", None)
        )

        self.assertIsNotNone(q_dataloader_params)
        self.assertIsNotNone(eval_dataloader_params)

        self.assertEqual(q_dataloader_params, expected_params)
        self.assertEqual(eval_dataloader_params, expected_params)


class TestEvaluationParser(TestParser):
    """Test evaluation part of parser."""

    def test_non_positive_cores_per_instance_fails(self) -> None:
        """Test parsing evaluation with non positive cores per instance."""
        with self.assertRaisesRegex(
            ClientErrorException,
            "At least one core per instance must be used.",
        ):
            self.parser.parse(
                {
                    "evaluation": {
                        "cores_per_instance": 0,
                    },
                },
            )

    @patch("neural_compressor.ux.components.configuration_wizard.configuration_parser.HWInfo")
    def test_too_big_num_cores_fails(self, mocked_hw_info: MagicMock) -> None:
        """Test parsing evaluation with non positive cores per instance."""
        mocked_hw_info.return_value.cores = 4
        with self.assertRaisesRegex(
            ClientErrorException,
            "Requested 10 cores per instance, while only 4 available.",
        ):
            self.parser.parse(
                {
                    "evaluation": {
                        "cores_per_instance": 10,
                    },
                },
            )

    def test_non_positive_instances_fails(self) -> None:
        """Test parsing evaluation with non positive cores per instance."""
        with self.assertRaisesRegex(
            ClientErrorException,
            "At least one instance must be used.",
        ):
            self.parser.parse(
                {
                    "evaluation": {
                        "cores_per_instance": 1,
                        "instances": 0,
                    },
                },
            )

    @patch("neural_compressor.ux.components.configuration_wizard.configuration_parser.HWInfo")
    def test_too_big_instances_fails(self, mocked_hw_info: MagicMock) -> None:
        """Test parsing evaluation with non positive cores per instance."""
        mocked_hw_info.return_value.cores = 12
        with self.assertRaisesRegex(
            ClientErrorException,
            "Attempted to use 4 instances, while only 3 allowed.",
        ):
            self.parser.parse(
                {
                    "evaluation": {
                        "cores_per_instance": 4,
                        "instances": 4,
                    },
                },
            )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main endpoint for GUI."""
import os
from functools import wraps
from threading import Thread
from typing import Any, Callable

from flask import Flask
from flask import Request as WebRequest
from flask import jsonify, request
from flask_cors import CORS
from flask_socketio import SocketIO
from werkzeug.wrappers import Response as WebResponse

from neural_compressor.ux.utils.exceptions import InternalException
from neural_compressor.ux.utils.logger import log
from neural_compressor.ux.web.communication import MessageQueue, Request
from neural_compressor.ux.web.configuration import Configuration
from neural_compressor.ux.web.router import Router
from neural_compressor.ux.web.service.response_generator import ResponseGenerator

app = Flask(__name__, static_url_path="")
socketio = SocketIO()
router = Router()

METHODS = ["GET", "POST"]

# Suppress TensorFlow messages
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


def run_server(configuration: Configuration) -> None:
    """Run webserver on specified scheme, address and port."""
    addr = configuration.server_address
    server_port = configuration.server_port
    gui_port = configuration.gui_port
    token = configuration.token

    cors_allowed_origins = f"{configuration.scheme}://{addr}:{gui_port}"

    app.config["JSON_SORT_KEYS"] = False
    app.secret_key = token
    CORS(app, origins=cors_allowed_origins)
    socketio.init_app(
        app,
        cors_allowed_origins=cors_allowed_origins,
        max_http_buffer_size=2000,
    )

    socketio.run(app, host=addr, port=server_port)


@app.after_request
def block_iframe(response: WebResponse) -> WebResponse:
    """Block iframe and set others CSP."""
    response.headers["X-Frame-Options"] = "DENY"
    response.headers[
        "Content-Security-Policy"
    ] = "frame-ancestors 'none'; font-src 'self'; img-src 'self'; script-src 'self'"
    response.headers["Access-Control-Max-Age"] = "-1"
    return response


@app.after_request
def block_sniffing(response: WebResponse) -> WebResponse:
    """Block MIME sniffing."""
    response.headers["X-Content-Type-Options"] = "nosniff"
    return response


def require_api_token(func: Callable) -> Any:
    """Validate authorization token."""

    @wraps(func)
    def check_token(*args: str, **kwargs: str) -> Any:
        """Validate that correct token was provided."""
        provided_token = request.headers.get(
            "Authorization",
            request.args.to_dict().get("token", None),
        )

        if not app.secret_key == provided_token:
            return (
                "Invalid token, please use the URL displayed by the server on startup",
                403,
            )

        return func(*args, **kwargs)

    return check_token


@app.route("/", methods=METHODS)
def root() -> Any:
    """Serve JS application index."""
    return app.send_static_file("index.html")


@app.route("/api/<path:subpath>", methods=METHODS)
@require_api_token
def handle_api_call(subpath: str) -> Any:
    """Handle API access."""
    try:
        parameters = build_parameters(subpath, request)
        response = router.handle(parameters)
        if isinstance(response, WebResponse):
            return response
        return jsonify(response.data)
    except Exception as err:
        if isinstance(err, InternalException):
            log.critical(err)
        return ResponseGenerator.from_exception(err)


@app.route("/api/<path:subpath>", methods=["OPTIONS"])
def allow_api_call(subpath: str) -> Any:
    """Allow for API access."""
    return "OK"


@app.errorhandler(404)
def page_not_found(e: Any) -> Any:
    """Serve JS application index when no static file found."""
    return app.send_static_file("index.html")


@app.after_request
def disable_cache(response: WebResponse) -> WebResponse:
    """Disable cache on all requests."""
    response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
    response.headers["Pragma"] = "no-cache"
    response.headers["Expires"] = "0"
    response.headers["Cache-Control"] = "public, max-age=0"
    return response


def build_parameters(endpoint: str, request: WebRequest) -> Request:
    """Build domain object from flask request."""
    data = request.get_json() if request.is_json else request.args.to_dict(flat=False)
    return Request(request.method, endpoint, data)


def web_socket_publisher(web_socket: SocketIO) -> None:
    """Send messages from queue via web-socket to GUI."""
    queue = MessageQueue()
    while True:
        message = queue.get()
        web_socket.emit(
            message.subject,
            {"status": message.status, "data": message.data},
            broadcast=True,
        )
        socketio.sleep(0)


publisher = Thread(
    target=web_socket_publisher,
    args=(socketio,),
)
publisher.daemon = True
publisher.start()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import numpy as np


class Tensor(object):
    def __init__(self,
                 name='',
                 source_op=[],
                 dest_op=[],
                 shape=None,
                 data=None,
                 dtype=None,
                 location=None):
        self._name = name
        # assume data in neural_compressor tensor should be numpy array
        # however, we don't assign the data diretly if the tensor is
        # const like weight when parse model
        # otherwise it will make a bloated new graph
        # but it still can be set when using the constructed new graph
        self._data = data
        self._shape = shape
        self._dtype = dtype
        # location in bin file if const
        self._location = location
        self._source_op = source_op
        self._dest_op = dest_op

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = shape

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        self._dtype = dtype

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        self._location = location

    @property
    def source_op(self):
        return self._source_op

    @source_op.setter
    def source_op(self, source_op):
        self._source_op = source_op

    @property
    def dest_op(self):
        return self._dest_op

    @dest_op.setter
    def dest_op(self, dest_op):
        self._dest_op = dest_op

    @property
    def config(self):
        conf_dict = OrderedDict()
        if self._dtype is not None:
            conf_dict['dtype'] = self._dtype
        if self._shape is not None:
            conf_dict['shape'] = self._shape
        if self._location is not None:
            conf_dict['location'] = self._location

        return conf_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .base_operator import QuantOperatorBase
from neural_compressor.adaptor.ox_utils.util import QuantizedValue

# For operators that support 8bits operations directly, and output could
# reuse input[0]'s type, zeropoint, scale; For example,Transpose, Reshape, etc.
class Direct8BitOp(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node

        # Quantize when input[0] is quantized already. Otherwise keep it.
        if node.input[0] not in self.quantizer.quantized_value_map:
            self.quantizer.new_nodes += [node]
            return
            
        quantized_input_value = self.quantizer.quantized_value_map[node.input[0]]
        quantized_output_value = QuantizedValue(node.output[0], node.output[0] + "_quantized",
                                                quantized_input_value.scale_name, 
                                                quantized_input_value.zp_name,
                                                quantized_input_value.value_type)
        self.quantizer.quantized_value_map[node.output[0]] = quantized_output_value
        node.input[0] = node.input[0] + "_quantized"
        node.output[0] = node.output[0] + "_quantized"

        self.quantizer.new_nodes += [node]



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph Optimization config test."""

import unittest

from neural_compressor.ux.utils.exceptions import ClientErrorException
from neural_compressor.ux.utils.workload.graph_optimization import GraphOptimization


class GraphOptimizationConfig(unittest.TestCase):
    """Graph Optimization config tests."""

    def __init__(self, *args: str, **kwargs: str) -> None:
        """Graph Optimization config test constructor."""
        super().__init__(*args, **kwargs)

    def test_graph_optimization_constructor(self) -> None:
        """Test Graph Optimization config constructor."""
        data = {
            "precisions": "bf16, fp32",
            "op_wise": {
                "weight": {
                    "granularity": "per_channel",
                    "scheme": "asym",
                    "dtype": "bf16",
                    "algorithm": "minmax",
                },
                "activation": {
                    "granularity": "per_tensor",
                    "scheme": "sym",
                    "dtype": "int8",
                    "algorithm": "minmax",
                },
            },
        }
        graph_optimization = GraphOptimization(data)

        self.assertEqual(graph_optimization.precisions, "bf16,fp32")
        self.assertIsNotNone(graph_optimization.op_wise)
        self.assertDictEqual(
            graph_optimization.op_wise,
            {
                "weight": {
                    "granularity": "per_channel",
                    "scheme": "asym",
                    "dtype": "bf16",
                    "algorithm": "minmax",
                },
                "activation": {
                    "granularity": "per_tensor",
                    "scheme": "sym",
                    "dtype": "int8",
                    "algorithm": "minmax",
                },
            },
        )

    def test_graph_optimization_constructor_defaults(self) -> None:
        """Test Graph Optimization config constructor defaults."""
        graph_optimization = GraphOptimization()

        self.assertIsNone(graph_optimization.precisions)
        self.assertIsNone(graph_optimization.op_wise)

    def test_set_precisions_string(self) -> None:
        """Test setting precisions in Graph Optimization config."""
        graph_optimization = GraphOptimization()
        graph_optimization.set_precisions(" bf16, fp32 ")
        self.assertEqual(graph_optimization.precisions, "bf16,fp32")

    def test_set_precisions_list(self) -> None:
        """Test setting precisions in Graph Optimization config."""
        graph_optimization = GraphOptimization()
        graph_optimization.set_precisions(["bf16", "fp32 ", " int8"])
        self.assertEqual(graph_optimization.precisions, "bf16,fp32,int8")

    def test_set_precisions_error(self) -> None:
        """Test overwriting precisions in Graph Optimization config."""
        graph_optimization = GraphOptimization()
        with self.assertRaises(ClientErrorException):
            graph_optimization.set_precisions(1)

    def test_graph_optimization_serializer(self) -> None:
        """Test Graph Optimization config serializer."""
        data = {
            "precisions": "bf16, fp32",
            "op_wise": {
                "weight": {
                    "granularity": "per_channel",
                    "scheme": "asym",
                    "dtype": "bf16",
                    "algorithm": "minmax",
                },
                "activation": {
                    "granularity": "per_tensor",
                    "scheme": "sym",
                    "dtype": "int8",
                    "algorithm": "minmax",
                },
            },
        }
        graph_optimization = GraphOptimization(data)
        result = graph_optimization.serialize()

        self.assertDictEqual(
            result,
            {
                "precisions": "bf16,fp32",
                "op_wise": {
                    "weight": {
                        "granularity": "per_channel",
                        "scheme": "asym",
                        "dtype": "bf16",
                        "algorithm": "minmax",
                    },
                    "activation": {
                        "granularity": "per_tensor",
                        "scheme": "sym",
                        "dtype": "int8",
                        "algorithm": "minmax",
                    },
                },
            },
        )

    def test_graph_optimization_serializer_defaults(self) -> None:
        """Test Graph Optimization config serializer."""
        graph_optimization = GraphOptimization()
        result = graph_optimization.serialize()

        self.assertDictEqual(result, {})


if __name__ == "__main__":
    unittest.main()



"""Tests for quantization"""
import numpy as np
import unittest
import os
import shutil
import yaml
import tensorflow as tf
import torch
import torchvision

def build_fake_yaml():
    fake_yaml = '''
        model:
          name: fake_yaml
          framework: tensorflow
          inputs: x
          outputs: op_to_store
        device: cpu
        evaluation:
          accuracy:
            metric:
              topk: 1
        tuning:
            strategy:
              name: mse
            accuracy_criterion:
              relative: 0.01
            workspace:
              path: saved
        '''
    y = yaml.load(fake_yaml, Loader=yaml.SafeLoader)
    with open('fake_yaml.yaml',"w",encoding="utf-8") as f:
        yaml.dump(y,f)
    f.close()

def build_fake_yaml2():
    fake_yaml = '''
        model:
          name: fake_yaml
          framework: tensorflow
          inputs: x
          outputs: op_to_store
        device: cpu
        evaluation:
          accuracy:
            metric:
              topk: 1
        tuning:
          strategy:
            name: mse
          exit_policy:
            max_trials: 5
          accuracy_criterion:
            relative: -0.01
          workspace:
            path: saved
        '''
    y = yaml.load(fake_yaml, Loader=yaml.SafeLoader)
    with open('fake_yaml2.yaml',"w",encoding="utf-8") as f:
        yaml.dump(y,f)
    f.close()

def build_ox_yaml():
    fake_yaml = '''
        model:
          name: fake_yaml
          framework: onnxrt_qlinearops
          inputs: input
          outputs: output
        evaluation:
          accuracy:
            metric:
              Accuracy: {}
        tuning:
            strategy:
              name: mse
            accuracy_criterion:
              relative: -0.01
              higher_is_better: False
            exit_policy:
              max_trials: 3
            workspace:
              path: saved
        '''
    y = yaml.load(fake_yaml, Loader=yaml.SafeLoader)
    with open('ox_yaml.yaml',"w",encoding="utf-8") as f:
        yaml.dump(y,f)
    f.close()

def build_fake_model():
    try:
        graph = tf.Graph()
        graph_def = tf.GraphDef()
        with tf.Session() as sess:
            x = tf.placeholder(tf.float64, shape=(1,3,3,1), name='x')
            y = tf.constant(np.random.random((2,2,1,1)), name='y')
            op = tf.nn.conv2d(input=x, filter=y, strides=[1,1,1,1], padding='VALID', name='op_to_store')

            sess.run(tf.global_variables_initializer())
            constant_graph = tf.graph_util.convert_variables_to_constants(sess, sess.graph_def, ['op_to_store'])

        graph_def.ParseFromString(constant_graph.SerializeToString())
        with graph.as_default():
            tf.import_graph_def(graph_def, name='')
    except:
        graph = tf.Graph()
        graph_def = tf.compat.v1.GraphDef()
        with tf.compat.v1.Session() as sess:
            x = tf.compat.v1.placeholder(tf.float64, shape=(1,3,3,1), name='x')
            y = tf.compat.v1.constant(np.random.random((2,2,1,1)), name='y')
            op = tf.nn.conv2d(input=x, filters=y, strides=[1,1,1,1], padding='VALID', name='op_to_store')

            sess.run(tf.compat.v1.global_variables_initializer())
            constant_graph = tf.compat.v1.graph_util.convert_variables_to_constants(sess, sess.graph_def, ['op_to_store'])

        graph_def.ParseFromString(constant_graph.SerializeToString())
        with graph.as_default():
            tf.import_graph_def(graph_def, name='')
    return graph

def build_ox_model():
    path = "mb_v2.onnx"
    model = torchvision.models.mobilenet_v2()

    x = torch.randn(100, 3, 224, 224, requires_grad=True)
    torch_out = model(x)

    torch.onnx.export(model,
                      x,
                      path,
                      export_params=True,
                      opset_version=12,
                      do_constant_folding=True,
                      input_names = ["input"],
                      output_names = ["output"],
                      dynamic_axes={"input" : {0 : "batch_size"},
                                    "output" : {0 : "batch_size"}})

class dataset:
    def __init__(self):
        self.data = []
        self.label = []
        for i in range(10):
            self.data.append(np.zeros((3, 224, 224)).astype(np.float32))
            self.label.append(0)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index], self.label[index]

class TestQuantization(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.constant_graph = build_fake_model()
        build_fake_yaml()
        build_fake_yaml2()
        build_ox_model()
        build_ox_yaml()

    @classmethod
    def tearDownClass(self):
        os.remove('fake_yaml.yaml')
        os.remove('fake_yaml2.yaml')
        os.remove('ox_yaml.yaml')
        os.remove('mb_v2.onnx')

        shutil.rmtree("saved", ignore_errors=True)

    def test_ru_mse_one_trial(self):
        from neural_compressor.experimental import Quantization, common
        quantizer = Quantization('fake_yaml.yaml')
        dataset = quantizer.dataset('dummy', (100, 3, 3, 1), label=True)
        quantizer.calib_dataloader = common.DataLoader(dataset)
        quantizer.eval_dataloader = common.DataLoader(dataset)
        quantizer.model = self.constant_graph
        quantizer()

    def test_ru_mse_max_trials(self):
        from neural_compressor.experimental import Quantization, common
        quantizer = Quantization('fake_yaml2.yaml')
        dataset = quantizer.dataset('dummy', (100, 3, 3, 1), label=True)
        quantizer.calib_dataloader = common.DataLoader(dataset)
        quantizer.eval_dataloader = common.DataLoader(dataset)
        quantizer.model = self.constant_graph
        quantizer()

    def test_ox_mse(self):
        from neural_compressor.experimental import Quantization, common
        quantizer = Quantization('ox_yaml.yaml')
        ds = dataset()
        quantizer.calib_dataloader = common.DataLoader(ds)
        quantizer.eval_dataloader = common.DataLoader(ds)
        quantizer.model = common.Model('mb_v2.onnx')
        quantizer()

if __name__ == "__main__":
    unittest.main()



import unittest
import tensorflow as tf
import numpy as np
import copy
from tensorflow.python.framework import graph_util
from neural_compressor.adaptor.tf_utils.quantize_graph.quantize_graph_common import QuantizeGraphHelper
from neural_compressor.adaptor.tf_utils.graph_rewriter.generic.fold_batch_norm import FoldBatchNormNodesOptimizer


class TestFoldBatchnorm(unittest.TestCase):
    tf.compat.v1.disable_eager_execution()
    x = tf.compat.v1.placeholder(tf.float32, [1, 224, 224, 3], name="input")
    conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 3, 32],
                                             initializer=tf.compat.v1.random_normal_initializer())
    conv_bias = tf.compat.v1.get_variable("bias", [32],
                                          initializer=tf.compat.v1.random_normal_initializer())
    beta = tf.compat.v1.get_variable(name='beta',
                                     shape=[32],
                                     initializer=tf.compat.v1.random_normal_initializer())
    gamma = tf.compat.v1.get_variable(name='gamma',
                                      shape=[32],
                                      initializer=tf.compat.v1.random_normal_initializer())
    conv1 = tf.nn.conv2d(x, conv_weights, strides=[1, 1, 1, 1], padding="SAME")
    conv_bias = tf.nn.bias_add(conv1, conv_bias)
    normed = tf.compat.v1.layers.batch_normalization(conv_bias)
    with tf.compat.v1.Session() as sess:
        sess.run(tf.compat.v1.global_variables_initializer())
        output_graph_def = graph_util.convert_variables_to_constants(
            sess=sess,
            input_graph_def=sess.graph_def,
            output_node_names=[normed.name.split(':')[0]])
        output_graph_def = QuantizeGraphHelper.remove_training_nodes(
            output_graph_def, protected_nodes=[normed.name.split(':')[0]])
        graph_def = copy.deepcopy(output_graph_def)
        fold_graph_def = FoldBatchNormNodesOptimizer(output_graph_def).do_transformation()

    def test_fold_output_values(self):
        input_data = np.random.randn(1, 224, 224, 3)
        graph = tf.compat.v1.Graph()
        fold_graph = tf.compat.v1.Graph()
        with graph.as_default():
            tf.compat.v1.import_graph_def(self.graph_def, name='')

        with tf.compat.v1.Session(graph=graph) as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            x = graph.get_tensor_by_name('input:0')
            normed = graph.get_tensor_by_name('batch_normalization/FusedBatchNormV3:0')
            y = sess.run(normed, feed_dict={x: input_data})

        with fold_graph.as_default():
            tf.compat.v1.import_graph_def(self.fold_graph_def, name='')
        with tf.compat.v1.Session(graph=fold_graph) as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            x = fold_graph.get_tensor_by_name('input:0')
            normed = fold_graph.get_tensor_by_name('batch_normalization/FusedBatchNormV3:0')
            y_fold = sess.run(normed, feed_dict={x: input_data})
        assert np.allclose(y, y_fold, rtol=1e-05, atol=1e-05)

    def test_do_transform(self):
        for node in self.fold_graph_def.node:
            assert node.op not in ["FusedBatchNormV3"]


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Execution common proc module."""

import datetime
import json
import os
import re
import subprocess
import uuid
from contextlib import ExitStack
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union

from neural_compressor.ux.utils.logger import log


class Proc(object):
    """Process class that runs commands from shell."""

    def __init__(
        self,
        output_dir: str = ".",
        pid: Optional[str] = None,
        request_id: Optional[str] = None,
        filename: Optional[Union[str, List[str]]] = None,
        additional_log_names: List[str] = [],
    ) -> None:
        """
        Initialize class parameters.

        :param output_dir: path to directory where process output saved
        :param pid unique aibt process identifier
        :return: the constructor returns no value
        """
        pid = pid if pid else uuid.uuid4().hex

        if filename:
            name = filename
        elif request_id:
            name = request_id
        else:
            name = pid
        log_names = [f"{name}.txt"]
        info_file_name = f"{name}.proc"

        log_names.extend(additional_log_names)

        if not os.path.exists(output_dir):
            try:
                os.makedirs(output_dir)
            except FileExistsError:
                log.debug("Directory %s already exists.", output_dir)

        # process output file
        self.stage_output_path = os.path.join(output_dir, log_names[0])
        self.log_paths = [os.path.join(output_dir, filename) for filename in log_names]

        # process info file
        self.proc_info_path = os.path.join(output_dir, info_file_name)

        # process information
        self.args: Optional[Iterable[Any]] = None
        self.return_code: Optional[int] = None
        self.time_start: Optional[datetime.datetime] = None
        self.time_stop: Optional[datetime.datetime] = None
        # default ignore exit code is only 0
        self.ignore_exit_codes = [0]

    def run(
        self,
        args: List[Any],
        executable: Optional[Any] = None,
        shell: bool = False,
        cwd: Optional[str] = None,
        env: Optional[dict] = None,
        universal_newlines: bool = False,
        startupinfo: Optional[Any] = None,
        creationflags: int = 0,
        ignore_exit_codes: Union[list, Any] = None,
    ) -> subprocess.Popen:
        """
        Execute call for process.

        :param args:
        :param executable:
        :param shell:
        :param cwd:
        :param env:
        :param universal_newlines:
        :param startupinfo:
        :param creationflags:
        :param ignore_exit_codes:
        :return: Popen
        """
        self.time_start = datetime.datetime.utcnow()
        self.time_stop = self.time_start
        self.return_code = None
        self.ignore_exit_codes = (
            ignore_exit_codes if ignore_exit_codes and isinstance(ignore_exit_codes, list) else [0]
        )

        try:
            self.args = args
            cmd: Union[str, Any] = " ".join(map(str, args)) if shell else map(str, args)
            proc = subprocess.Popen(
                cmd,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=shell,  # nosec
                executable=executable,
                cwd=cwd,
                env=env,
                universal_newlines=universal_newlines,
                startupinfo=startupinfo,
                creationflags=creationflags,
            )
            with ExitStack() as stack:
                files = [
                    stack.enter_context(open(fname, "a", encoding="utf-8"))
                    for fname in self.log_paths
                ]
                for line in proc.stdout:  # type: ignore
                    decoded_line = line.decode("utf-8", errors="ignore").strip()
                    log.debug(decoded_line)
                    for log_file in files:
                        log_file.write(decoded_line + "\n")
                        log_file.flush()
            proc.wait()
            self.time_stop = datetime.datetime.utcnow()
            self.return_code = proc.returncode
            msg = "Exit code: {}".format(self.return_code)
            if self.return_code in self.ignore_exit_codes:
                log.debug(msg)
            else:
                log.error(msg)
                log.critical("...")
                for line in self.tail:
                    log.critical(f"\t{line.strip()}")
                relative_output_path = re.sub(
                    r"^.*stages/",
                    "stages/",
                    self.output_path,
                )
                log.critical(f"\tMore in file: {relative_output_path}")
                log.critical("...")

            return proc

        finally:
            self.__save_proc_info()

    def __save_proc_info(self) -> bool:
        """
        Save all proc information in .proc file.

        :raise: bool
        :return: True if save was success and False in all other cases
        """
        if self.time_start and self.time_stop:
            proc_info = {
                "time_start": self.time_start.isoformat(),
                "time_stop": self.time_stop.isoformat(),
                "cmd": self.args,  # list of arguments
                "output_path": self.output_path,
            }

        with open(self.proc_info_path, mode="w") as proc_info_file:
            json.dump(proc_info, proc_info_file, sort_keys=True, indent=4)
            return True
        return False

    @property
    def output(self) -> Iterator[str]:
        """
        Property returns lines form execution.

        :return: Iterator with output lines from process execution
        """
        with open(self.stage_output_path, encoding="utf-8") as log_file:
            for line in log_file:
                yield line

    @property
    def info(self) -> Dict[str, Any]:
        """
        Info contains information about process arguments start and stop time.

        Example:
            "cmd": [
                "docker",
                "pull",
                "hub.docker.intel.com/aibt_tensorflow/centos-7_3_mkldnn_tensorflow:latest"
            ],
            "output_path": "111224/stages/pull_image/0c750a01-90e0-4aa8-99fe-a194d55a2046.txt",
            "time_start": "2018-05-17T14:47:22.352502",
            "time_stop": "2018-05-17T14:47:25.110410"

        :return: Process information
        """
        with open(self.proc_info_path, encoding="utf-8") as proc_info_file:
            return json.load(proc_info_file)

    @property
    def info_path(self) -> str:
        """
        Process .proc file with runtime process information.

        :return: absolute path to process .proc file
        """
        return self.proc_info_path

    @property
    def output_path(self) -> str:
        """
        Process .txt file with process logs.

        :return: absolute path to process .txt file
        """
        if os.path.isfile(self.stage_output_path):
            return self.stage_output_path

        raise Exception("Missing output path in {}".format(self.stage_output_path))

    @property
    def tail(self) -> list:
        """
        Property returns lines form execution along open file.

        :return: Lines from process execution
        """
        lines = 20
        buffer = 4098
        # place holder for the lines found
        lines_found: list = []

        # block counter will be multiplied by buffer
        # to get the block size from the end
        block_counter = -1

        with open(self.output_path, "r") as f:
            # loop until we find X lines
            while len(lines_found) < lines:
                try:
                    f.seek(block_counter * buffer, os.SEEK_END)
                except IOError:  # either file is too small, or too many lines requested
                    f.seek(0)
                    lines_found = f.readlines()
                    break

                lines_found = f.readlines()
                block_counter -= 1

        return lines_found[-lines:]

    @property
    def is_ok(self) -> bool:
        """
        Check if process execute as expected.

        :return: bool if return code from cmd is in ignore_exit_codes return True, False if not
        """
        try:
            return self.return_code in self.ignore_exit_codes
        except Exception:
            return False

    def remove_logs(self) -> None:
        """
        Remove logs generated by process.

        :return: Nothing
        """
        os.remove(self.output_path)
        os.remove(self.info_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from tensorflow.python.framework import tensor_util
from tensorflow.core.framework import attr_value_pb2
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes

from ..graph_base import GraphRewriterBase
from ..graph_util import GraphAnalyzer
from ..graph_util import GraphRewriterHelper as Helper


class FuseConvRequantizeTransformer(GraphRewriterBase):
    """Fuse Quantized Conv Op with the successor Requantize Op.
    """
    fuse_patterns = [[
        "QuantizedConv2DWithBiasAndRelu",
        "QuantizedDepthwiseConv2DWithBiasAndRelu",
        "QuantizedConv2DWithBias",
    ], ['RequantizePerChannel', 'Requantize']]
    sum_pattern = [["QuantizedConv2DWithBiasSumAndRelu", "QuantizedConv2DWithBiasReluAndSum"],
                     ['RequantizePerChannel', 'Requantize']]

    def __init__(self, model, device='cpu'):
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

    def do_transformation(self):
        """Fuse the quantized op with the following requantize op.
            The transformation has two stages, the first step is to fuse the patterns
            defined in self.fuse_patterns and the last step is to fuse the self.sum_patterns.
        Returns:
            [graphdef]: the optimized graphdef object
        """
        int8_type = dtypes.qint8.as_datatype_enum
        uint8_type = dtypes.quint8.as_datatype_enum
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum

        while True:
            target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.fuse_patterns)
            if len(target_nodes) == 0:
                break

            i = target_nodes[0]

            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndRequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if 'Tinput' in quantized_node.attr:
                new_node.attr["Tinput"].CopyFrom(quantized_node.attr['Tinput'])
            if 'Tfilter' in quantized_node.attr:
                new_node.attr["Tfilter"].CopyFrom(quantized_node.attr['Tfilter'])
            if 'strides' in quantized_node.attr:
                new_node.attr["strides"].CopyFrom(quantized_node.attr['strides'])
            if 'padding' in quantized_node.attr:
                new_node.attr["padding"].CopyFrom(quantized_node.attr['padding'])
            if 'alpha' in quantized_node.attr:
                new_node.attr["alpha"].CopyFrom(quantized_node.attr['alpha'])

            parent_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = self.graph_info[new_node.input[6]].node
            min_filter_node = self.graph_info[new_node.input[5]].node
            last_node = self.graph_info[new_node.input[0]].node
            if last_node.op.find('Requantize') != -1:
                bias_node = self.graph_info[new_node.input[2]].node
                max_input_node = self.graph_info[last_node.input[-1]].node
                min_input_node = self.graph_info[last_node.input[-2]].node
                min_input = (min_input_node.attr['value'].tensor.float_val)[0]
                max_input = (max_input_node.attr['value'].tensor.float_val)[0]
                if 'Depthwise' in quantized_node_op or requantize_node.op.find('PerChannel') != -1:
                    channel_size = max_filter_node.attr['value'].tensor.tensor_shape.dim[0].size
                    max_filter_tensor = tensor_util.MakeNdarray(
                        min_filter_node.attr['value'].tensor)
                    min_filter_tensor = tensor_util.MakeNdarray(
                        min_filter_node.attr['value'].tensor)
                else:
                    channel_size = 1
                    max_filter_tensor = []
                    min_filter_tensor = []
                    max_filter_tensor.append((max_filter_node.attr['value'].tensor.float_val)[0])
                    min_filter_tensor.append((min_filter_node.attr['value'].tensor.float_val)[0])
                bias_tensor = tensor_util.MakeNdarray(
                    self.graph_info[new_node.input[2]].node.attr['value'].tensor)

                activation_range = 127.0 if new_node.attr["Tinput"].type == dtypes.qint8 else 255.0

                int32_bias = Helper.generate_int32_bias_for_conv(
                    bias_tensor, channel_size, max_input, min_input,
                    max_filter_tensor, min_filter_tensor, activation_range)
                bias_node.attr['dtype'].CopyFrom(
                    attr_value_pb2.AttrValue(
                        type=float32_type if self.device == 'gpu' else qint32_type))
                bias_node.attr['value'].CopyFrom(
                    attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(
                        bias_tensor if self.device == 'gpu' else int32_bias, dtypes.
                        float32 if self.device == 'gpu' else dtypes.int32, bias_tensor.shape)))

                bias_node.attr['value'].tensor.dtype = float32_type \
                                        if self.device == 'gpu' else qint32_type
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type \
                                                if self.device == 'gpu' else qint32_type))
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            if "padding_list" in quantized_node.attr:
                new_node.attr["padding_list"].CopyFrom(quantized_node.attr['padding_list'])
            if "dilations" in quantized_node.attr:
                new_node.attr["dilations"].CopyFrom(quantized_node.attr['dilations'])

            if quantized_node.op == "QuantizedConv2D" or \
                    quantized_node.op == "QuantizedConv2DWithBias" or \
                    'alpha' in quantized_node.attr:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=int8_type))
            else:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
            self.graph_analyzer.replace_single_node(
                new_node, [parent_node_name], quantized_node_name,
                [self.graph_info[requantize_node_name].outputs[0]], requantize_node_name)
            self.graph_analyzer.remove_node(quantized_node_name)

        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.sum_pattern)
        while target_nodes:
            i = target_nodes[0]
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndRequantize"
            new_node.name = requantize_node_name

            for _, value in enumerate(quantized_node.input[:-1]):
                new_node.input.append(value)

            new_node.attr["Tinput"].CopyFrom(quantized_node.attr['Tinput'])
            new_node.attr["Tfilter"].CopyFrom(quantized_node.attr['Tfilter'])
            new_node.attr["strides"].CopyFrom(quantized_node.attr['strides'])
            new_node.attr["padding"].CopyFrom(quantized_node.attr['padding'])

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            deq_node = self.graph_info[Helper.node_name_from_input(quantized_node.input[-1])].node
            if deq_node.op != 'Dequantize' or deq_node.op.find("Quantize") != -1:
                target_nodes.remove(i)
                continue
            if deq_node.op == 'Dequantize':
                original_summand_node = self.graph_info[Helper.node_name_from_input(
                    deq_node.input[0])].node
            else:
                original_summand_node = deq_node
            summand_op_type = uint8_type if dtypes.as_dtype(
                deq_node.attr["T"].type) == uint8_type else int8_type

            for j in range(3):
                new_node.input.append(original_summand_node.name + ':{}'.format(j))

            if "padding_list" in quantized_node.attr:
                new_node.attr["padding_list"].CopyFrom(quantized_node.attr['padding_list'])

            if "dilations" in quantized_node.attr:
                new_node.attr["dilations"].CopyFrom(quantized_node.attr['dilations'])
            if "alpha" in quantized_node.attr:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=int8_type))
            else:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))

            new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            if quantized_node_op == 'QuantizedConv2DWithBiasReluAndSum':
                new_node.op = 'QuantizedConv2DWithBiasReluAndSumAndRequantize'
                if "alpha" in quantized_node.attr:
                    new_node.attr["alpha"].CopyFrom(quantized_node.attr['alpha'])

            elif summand_op_type == int8_type:
                new_node.op = "QuantizedConv2DWithBiasSignedSumAndReluAndRequantize"

            new_node.attr["Tsummand"].CopyFrom(attr_value_pb2.AttrValue(type=summand_op_type))

            self.graph_analyzer.replace_single_node(
                new_node, [quantized_node.input[0], original_summand_node.name],
                quantized_node.name, self.graph_info[requantize_node_name].outputs,
                requantize_node_name)
            self.graph_analyzer.remove_node(quantized_node_name)

            if deq_node.op == 'Dequantize':
                self.graph_analyzer.remove_node_with_single_input_output(deq_node.name)
            target_nodes.remove(i)

        return self.graph_analyzer.dump_graph()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tuning history."""
import os.path
from time import sleep
from typing import Optional

from neural_compressor.utils.utility import get_tuning_history
from neural_compressor.ux.utils.exceptions import NotFoundException
from neural_compressor.ux.utils.workload.config import Config
from neural_compressor.ux.web.service.history_snapshot_parser import HistorySnapshotParser


def tuning_history(workload_id: str) -> dict:
    """Get tuning history for requested Workload."""
    from neural_compressor.ux.web.service.workload import WorkloadService

    workload_data = WorkloadService.get_workload_data_by_id(workload_id)

    response = {
        "workload_id": workload_data.get("id"),
    }

    tuning_data = _build_tuning_history(workload_data)
    response.update(tuning_data)

    return response


def _build_tuning_history(workload_data: dict) -> dict:
    """Build tuning history data response."""
    history_path = tuning_history_filename(workload_data.get("id", ""))
    if not os.path.isfile(history_path):
        raise NotFoundException(f"Unable to find tuning history file {history_path}")
    history_snapshot = get_tuning_history(history_path)

    config_path = workload_data.get("config_path")
    if not config_path:
        raise NotFoundException("Unable to find config file")
    config = Config()
    config.load(config_path)
    objective = config.tuning.objective

    history_snapshot_parser = HistorySnapshotParser(
        history_snapshot,
        "performance" == objective,
    )
    tuning_data = history_snapshot_parser.parse_history_snapshot()

    if not tuning_data.get("baseline_accuracy"):
        raise NotFoundException("Can't find baseline accuracy")

    baseline_accuracy = tuning_data.get("baseline_accuracy", 0)
    accuracy_criterion = config.tuning.accuracy_criterion
    if accuracy_criterion.relative:
        accuracy_criterion_type = "relative"
        accuracy_criterion_value = accuracy_criterion.relative
        minimal_accuracy = baseline_accuracy * (1 - accuracy_criterion_value)
    elif accuracy_criterion.absolute:
        accuracy_criterion_type = "absolute"
        accuracy_criterion_value = accuracy_criterion.absolute
        minimal_accuracy = baseline_accuracy - accuracy_criterion_value
    else:
        raise NotFoundException("Unknown accuracy type")

    response = {
        "accuracy_criterion_type": accuracy_criterion_type,
        "accuracy_criterion_value": accuracy_criterion_value,
        "minimal_accuracy": minimal_accuracy,
    }
    response.update(tuning_data)
    return response


def tuning_history_filename(workload_id: str) -> str:
    """Build tuning history filename based on id."""
    from neural_compressor.ux.web.service.workload import WorkloadService

    workload_data = WorkloadService.get_workload_data_by_id(workload_id)
    workload_path = workload_data.get("workload_path", "")
    history_path = os.path.join(workload_path, "history.snapshot")

    return history_path


class Watcher:
    """Tuning history watcher that sends update on file change."""

    def __init__(self, workload_id: str) -> None:
        """Initialize object."""
        self.workload_id = workload_id
        self.watch = False
        self.tuning_history_filename = tuning_history_filename(workload_id)
        self.last_tuning_history_timestamp = self.history_file_modification_time()

    def stop(self) -> None:
        """Signal watcher to stop."""
        self.watch = False

    def __call__(self) -> None:
        """Execute the watch."""
        from neural_compressor.ux.web.service.workload import WorkloadService

        self.watch = True
        while self.watch:
            if self.was_history_file_changed():
                WorkloadService.send_history_snapshot(self.workload_id)
            sleep(10)

    def was_history_file_changed(self) -> bool:
        """Check if history file was changed since last check."""
        current_tuning_history_file_timestamp = self.history_file_modification_time()
        if current_tuning_history_file_timestamp != self.last_tuning_history_timestamp:
            self.last_tuning_history_timestamp = current_tuning_history_file_timestamp
            return True
        return False

    def history_file_modification_time(self) -> Optional[float]:
        """Get modification date of history file."""
        try:
            return os.path.getmtime(self.tuning_history_filename)
        except OSError:
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workload service."""
from neural_compressor.ux.components.optimization.tuning_history import tuning_history
from neural_compressor.ux.utils.exceptions import NotFoundException
from neural_compressor.ux.utils.templates.workdir import Workdir
from neural_compressor.ux.web.communication import MessageQueue
from neural_compressor.ux.web.service.request_data_processor import RequestDataProcessor
from neural_compressor.ux.web.service.response_generator import Response, ResponseGenerator

mq = MessageQueue()


class WorkloadService:
    """Workload related services."""

    @staticmethod
    def get_config(data: dict) -> Response:
        """Get config file for requested Workload."""
        config_path = WorkloadService._get_workload_data_from_input_data(data).get("config_path")

        if not config_path:
            raise NotFoundException("Unable to find config file")

        return ResponseGenerator.serve_from_filesystem(
            path=config_path,
            mimetype="text/vnd.yaml",
        )

    @staticmethod
    def get_code_template(data: dict) -> Response:
        """Get code template file for requested Workload."""
        code_template_path = WorkloadService._get_workload_data_from_input_data(data).get(
            "code_template_path",
        )

        if not code_template_path:
            raise NotFoundException("Unable to find code template file")

        return ResponseGenerator.serve_from_filesystem(
            path=code_template_path,
            mimetype="text/x-python",
        )

    @staticmethod
    def get_output(data: dict) -> Response:
        """Get config file for requested Workload."""
        log_path = WorkloadService._get_workload_data_from_input_data(data).get("log_path")

        if not log_path:
            raise NotFoundException("Unable to find output log")

        try:
            response = ResponseGenerator.serve_from_filesystem(
                path=log_path,
                mimetype="text/plain",
            )
        except Exception as e:
            response = ResponseGenerator.from_exception(e)

        return ResponseGenerator.add_refresh(
            response=response,
            refresh_time=3,
        )

    @staticmethod
    def request_history_snapshot(data: dict) -> None:
        """Get tuning history for requested Workload."""
        workload_id = RequestDataProcessor.get_string_value(data, "workload_id")
        WorkloadService.send_history_snapshot(workload_id)

    @staticmethod
    def send_history_snapshot(workload_id: str) -> None:
        """Get tuning history for requested Workload."""
        try:
            response = tuning_history(workload_id)
            mq.post_success("tuning_history", response)
        except NotFoundException:
            mq.post_error(
                "tuning_history",
                {
                    "workload_id": workload_id,
                },
            )

    @staticmethod
    def _get_workload_data_from_input_data(data: dict) -> dict:
        """Return data for requested Workload."""
        workload_id = RequestDataProcessor.get_string_value(data, "workload_id")
        return WorkloadService.get_workload_data_by_id(workload_id)

    @staticmethod
    def get_workload_data_by_id(workload_id: str) -> dict:
        """Return data for requested Workload."""
        workdir = Workdir()
        workload_data = workdir.get_workload_data(workload_id)
        if not workload_data:
            raise NotFoundException(f"Unable to find workload with id: {workload_id}")
        return workload_data



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Saving Workload test."""
import unittest
from collections import OrderedDict
from typing import List
from unittest.mock import patch

from neural_compressor.ux.components.configuration_wizard.save_workload import (
    change_performance_dataloader_to_dummy_if_possible,
    get_height_width_from_size,
    get_shape_from_transforms,
)
from neural_compressor.ux.components.model.shape import Shape
from neural_compressor.ux.utils.exceptions import NotFoundException
from neural_compressor.ux.utils.workload.config import Config
from neural_compressor.ux.utils.workload.dataloader import Dataset, Transform


class TestUpdateConfigWithDummy(unittest.TestCase):
    """Test updating config with Dummy dataloader."""

    def setUp(self) -> None:
        """Prepare environment."""
        super().setUp()

        model_repository_patcher = patch(
            "neural_compressor.ux.components.configuration_wizard.save_workload.ModelRepository",
        )
        self.addCleanup(model_repository_patcher.stop)
        model_repository_mock = model_repository_patcher.start()

        class MockedModelWithShapeOnly:
            """Totally faked Model class with only one property."""

            def __init__(self, shape: Shape):
                """Initialize object."""
                self.input_shape = shape
                self.shape_elements_order = ["height", "width", "channels"]

        def mocked_get_model(path: str) -> MockedModelWithShapeOnly:
            """Fake getting model based on path."""
            if "trusted" in path:
                shape = Shape(shape="1,2,3,4,5", trusted=True)
            else:
                shape = Shape(shape="6,7,8,9,0", trusted=False)
            return MockedModelWithShapeOnly(shape)

        model_repository_mock.return_value.get_model = mocked_get_model

    def test_get_shape_from_transforms_for_empty_list(self) -> None:
        """Test getting shape from transforms."""
        transforms: List[Transform] = []
        with self.assertRaisesRegex(NotFoundException, "Unable to detect shape for Dummy dataset"):
            get_shape_from_transforms(transforms, ["height", "width", "channels"])

    def test_get_shape_from_transforms_for_list_of_unknown_transforms(self) -> None:
        """Test getting shape from transforms."""
        transforms = [
            Transform("foo"),
            Transform("bar"),
            Transform("baz"),
        ]
        with self.assertRaisesRegex(NotFoundException, "Unable to detect shape for Dummy dataset"):
            get_shape_from_transforms(transforms, ["height", "width", "channels"])

    def test_get_shape_from_transforms_with_size_parameter(self) -> None:
        """Test getting shape from transforms."""
        for transform_name in [
            "Resize",
            "CenterCrop",
            "RandomResizedCrop",
            "RandomCrop",
            "CropResize",
        ]:
            transforms = [
                Transform("foo"),
                Transform(transform_name, {"size": [10, 20]}),
                Transform("baz"),
            ]
            expected = [10, 20, 3]
            actual = get_shape_from_transforms(transforms, ["height", "width", "channels"])
            self.assertEqual(expected, actual, f"Incorrect shape for {transform_name}")

    def test_get_shape_from_transforms_with_height_and_width_parameters(self) -> None:
        """Test getting shape from transforms."""
        for transform_name in [
            "ResizeCropImagenet",
            "BilinearImagenet",
        ]:
            transforms = [
                Transform("foo"),
                Transform(transform_name, {"height": 10, "width": 20}),
                Transform("baz"),
            ]
            expected = [10, 20, 3]
            actual = get_shape_from_transforms(transforms, ["height", "width", "channels"])
            self.assertEqual(expected, actual, f"Incorrect shape for {transform_name}")

    def test_get_shape_from_CropToBoundingBox(self) -> None:
        """Test getting shape from transforms."""
        transforms = [
            Transform("foo"),
            Transform("CropToBoundingBox", {"target_height": 10, "target_width": 20}),
            Transform("baz"),
        ]
        expected = [10, 20, 3]
        actual = get_shape_from_transforms(transforms, ["height", "width", "channels"])
        self.assertEqual(expected, actual)

    def test_get_shape_from_transforms_with_transposing(self) -> None:
        """Test getting shape from transforms."""
        transforms = [
            Transform("foo"),
            Transform("Resize", {"size": [10, 20]}),
            Transform("baz"),
        ]
        expected = [10, 20, 3]
        actual = get_shape_from_transforms(transforms, ["height", "width", "channels"])
        self.assertEqual(expected, actual, "Incorrect initial shape before transposing")

        rotate_right_transpose = Transform("Transpose", {"perm": [2, 0, 1]})

        # add first transpose
        transforms.append(rotate_right_transpose)
        expected = [3, 10, 20]
        actual = get_shape_from_transforms(transforms, ["height", "width", "channels"])
        self.assertEqual(expected, actual, "Incorrect shape after first transpose")

        # add second transpose
        transforms.append(rotate_right_transpose)
        expected = [20, 3, 10]
        actual = get_shape_from_transforms(transforms, ["height", "width", "channels"])
        self.assertEqual(expected, actual, "Incorrect shape after second transpose")

        # add third transpose
        transforms.append(rotate_right_transpose)
        expected = [10, 20, 3]
        actual = get_shape_from_transforms(transforms, ["height", "width", "channels"])
        self.assertEqual(expected, actual, "Incorrect shape after third transpose")

    def test_get_shape_fails_on_incorrect_transpose(self) -> None:
        """Test getting shape from transforms."""
        transforms = [
            Transform("foo"),
            Transform("Transpose"),
            Transform("baz"),
        ]
        with self.assertRaisesRegex(ValueError, "Unknown value of 'perm' argument in Transpose"):
            get_shape_from_transforms(transforms, ["height", "width", "channels"])

    def test_get_shape_from_different_resize_params(self) -> None:
        """Test getting shape from transforms."""
        sizes = {
            "int": {
                "size": 20,
                "expected": (20, 20),
            },
            "list - one element": {
                "size": [10],
                "expected": (10, 10),
            },
            "list - two elements": {
                "size": [10, 20],
                "expected": (10, 20),
            },
            "list - many elements": {
                "size": [10, 20, 30],
                "expected": (None, None),
            },
            "other": {
                "size": {"foo": "bar"},
                "expected": (None, None),
            },
        }

        for size_name, size_definition in sizes.items():
            expected = size_definition["expected"]  # type: ignore
            actual = get_height_width_from_size(size_definition["size"])  # type: ignore
            self.assertEqual(expected, actual, f"Incorrect size detection for {size_name}")

    def test_does_nothing_when_incorrect_domain(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("NLP", config)
        actual = config.serialize()

        self.assertEqual(expected, actual, "Config should not be changed by not supported domain")

    def test_does_nothing_when_evaluation_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.evaluation = None
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_evaluation_performance_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.evaluation.performance = None  # type: ignore  # type: ignore
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_evaluation_performance_dataloader_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.evaluation.performance.dataloader = None  # type: ignore
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_evaluation_performance_dataloader_dataset_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.evaluation.performance.dataloader.dataset = None  # type: ignore
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_quantization_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.quantization = None
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_quantization_calibration_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.quantization.calibration = None  # type: ignore
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_quantization_calibration_dataloader_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.quantization.calibration.dataloader = None  # type: ignore
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_quantization_calibration_dataloader_transform_empty(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.quantization.calibration.dataloader.transform.clear()  # type: ignore
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_does_nothing_when_unable_to_guess_shapes(
        self,
    ) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        config.quantization.calibration.dataloader.transform = OrderedDict(  # type: ignore
            {"foo": Transform("foo")},
        )
        expected = config.serialize()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)
        actual = config.serialize()

        self.assertEqual(expected, actual)

    def test_change_performance_dataloader_to_dummy(self) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)

        self.assertEqual(
            OrderedDict(),
            config.evaluation.performance.dataloader.transform,  # type: ignore
        )

        expected_dataset = Dataset(
            "dummy_v2",
            {
                "input_shape": [10, 20, 3],
                "label_shape": [1],
            },
        )
        self.assertEqual(
            expected_dataset.serialize(),
            config.evaluation.performance.dataloader.dataset.serialize(),  # type: ignore
        )

    def test_change_performance_dataloader_to_dummy_when_trusted_shape_detected(self) -> None:
        """Test change_performance_dataloader_to_dummy_if_possible."""
        config = self.get_eligible_config()
        config.model_path = "some model with trusted shape"

        change_performance_dataloader_to_dummy_if_possible("image_recognition", config)

        self.assertEqual(
            OrderedDict(),
            config.evaluation.performance.dataloader.transform,  # type: ignore
        )

        expected_dataset = Dataset(
            "dummy_v2",
            {
                "input_shape": [1, 2, 3, 4, 5],
                "label_shape": [1],
            },
        )
        self.assertEqual(
            expected_dataset.serialize(),
            config.evaluation.performance.dataloader.dataset.serialize(),  # type: ignore
        )

    def get_eligible_config(self) -> Config:
        """Build a Config that should be updated to Dummy."""
        return Config(
            {
                "evaluation": {
                    "performance": {
                        "dataloader": {
                            "dataset": {
                                "foo": {},
                            },
                        },
                    },
                },
                "quantization": {
                    "calibration": {
                        "dataloader": {
                            "dataset": {
                                "foo": {},
                            },
                            "transform": {
                                "Resize": {
                                    "size": [10, 20],
                                },
                            },
                        },
                    },
                },
            },
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy
import onnx
from onnxruntime.quantization.quant_utils import QuantizedValueType, \
                                                 attribute_to_kwarg
from .base_operator import QuantOperatorBase
from neural_compressor.adaptor.ox_utils.util import QuantizedValue

class QPad(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        assert (node.op_type == "Pad")

        # Only after version 11, it has the optional constant_value
        # If input[0] is not quantized, do not quanitize this node
        if (self.quantizer.opset_version < 11) or (node.input[0] not \
                                               in self.quantizer.quantized_value_map):
            super().quantize()
            return
        quantized_input_value = self.quantizer.quantized_value_map[node.input[0]]

        kwargs = {}
        for attribute in node.attribute:
            kv = attribute_to_kwarg(attribute)
            kwargs.update(kv)

        if 'mode' not in kwargs or kwargs['mode'] == b'constant':
            if len(node.input) > 2:  # There is 3rd input 'constant_value'
                zp_tensor = self.quantizer.model.get_initializer(quantized_input_value.zp_name)
                scale_tensor = \
                            self.quantizer.model.get_initializer(quantized_input_value.scale_name)
                # if zp_tensor is None or scale_tensor is None:
                #     super().quantize()
                #     return

                padding_constant_initializer = self.quantizer.model.get_initializer(node.input[2])
                if padding_constant_initializer is not None:
                    zp_array = onnx.numpy_helper.to_array(zp_tensor)
                    zp_value = zp_array.item() if zp_array.ndim == 0 else zp_array[0]
                    scale_array = onnx.numpy_helper.to_array(scale_tensor)
                    scale_value = scale_array.item() if scale_array.ndim == 0 else scale_array[0]
                    padding_constant_array = \
                                          onnx.numpy_helper.to_array(padding_constant_initializer)
                    quantized_padding_constant_array = quantize_nparray(
                                                    self.activation_dtype,
                                                    padding_constant_array, scale_value, zp_value)
                    quantized_padding_constant_name = node.input[2] + "_quantized"
                    quantized_padding_constant_initializer = onnx.numpy_helper.from_array(
                        quantized_padding_constant_array, quantized_padding_constant_name)
                    # Suppose this padding constant initializer only used by the node
                    self.quantizer.model.remove_initializer(padding_constant_initializer)
                    self.quantizer.model.add_initializer(quantized_padding_constant_initializer)
                    node.input[2] = quantized_padding_constant_name
                else:
                    pad_value_qnodes = self.quantizer._get_quantize_input_nodes(node, 2, 
                                             self.activation_dtype, self.activation_scheme)
                    self.quantizer.new_nodes += pad_value_qnodes
                    node.input[2] = pad_value_qnodes[0].output[0]
            else:
                # pad zero_point for original zero
                node.input.extend([quantized_input_value.zp_name])

        # Create an entry for output quantized value
        quantized_output_value = QuantizedValue(node.output[0], node.output[0] + "_quantized",
                                                quantized_input_value.scale_name, 
                                                quantized_input_value.zp_name,
                                                QuantizedValueType.Input)
        self.quantizer.quantized_value_map[node.output[0]] = quantized_output_value

        node.name = node.name + "_quant" if node.name != "" else ""
        node.input[0] = quantized_input_value.q_name
        node.output[0] = quantized_output_value.q_name
        self.quantizer.new_nodes += [node]

def quantize_nparray(qtype, arr, scale, zero_point, low=None, high=None):
    dtype = numpy.uint8 if qtype == "uint8" else numpy.int8
    cliplow = max(0 if dtype == numpy.uint8 else -127, -127 if low is None else low)
    cliphigh = min(255 if dtype == numpy.uint8 else 127, 255 if high is None else high)
    arr_fp32 = numpy.asarray((arr.astype(numpy.float32) / scale).round() + zero_point)
    numpy.clip(arr_fp32, cliplow, cliphigh, out=arr_fp32)
    return arr_fp32.astype(dtype)



#
#  -*- coding: utf-8 -*-
#
import unittest
import os
import yaml
import numpy as np
import tensorflow as tf

from neural_compressor.adaptor.tf_utils.quantize_graph.quantize_graph_for_intel_cpu import QuantizeGraphForIntel
from neural_compressor.adaptor.tf_utils.graph_rewriter.generic.strip_unused_nodes import StripUnusedNodesOptimizer
from neural_compressor.adaptor.tf_utils.graph_rewriter.generic.fold_batch_norm import FoldBatchNormNodesOptimizer
from tensorflow.python.framework import graph_util
from neural_compressor.adaptor.tensorflow import TensorflowQuery
from neural_compressor.adaptor.tf_utils.util import disable_random


def build_fake_yaml():
    fake_yaml = '''
        model:
          name: fake_yaml
          framework: tensorflow
          inputs: input
        device: cpu
        quantization:
          model_wise:
            weight:
                granularity: per_tensor
                scheme: sym
                dtype: int8
                algorithm: minmax
        evaluation:
          accuracy:
            metric:
              topk: 1
        tuning:
            strategy:
              name: basic
            accuracy_criterion:
              relative: 0.1
            exit_policy:
              performance_only: True
            workspace:
              path: saved
        '''
    y = yaml.load(fake_yaml, Loader=yaml.SafeLoader)
    with open('fake_yaml.yaml', "w", encoding="utf-8") as f:
        yaml.dump(y, f)
    f.close()


class TestConvBiasAddAddReluFusion(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        build_fake_yaml()

    @classmethod
    def tearDownClass(self):
        os.remove('fake_yaml.yaml')

    @disable_random()
    def test_conv_relu_fusion(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        top_relu = tf.nn.relu(x)
        paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        x_pad = tf.pad(top_relu, paddings, "CONSTANT")
        conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 16, 16],
                                                 initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(x_pad, conv_weights, strides=[1, 2, 2, 1], padding="VALID")
        relu = tf.nn.relu(conv)

        relu6 = tf.nn.relu6(relu, name='op_to_store')

        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])
            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization('fake_yaml.yaml')
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.eval_dataloader = common.DataLoader(dataset)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = output_graph_def
            output_graph = quantizer()
            found_conv_fusion = True

            for i in output_graph.graph_def.node:
                if i.op == 'Relu':
                    found_conv_fusion = False
                    break

            self.assertEqual(found_conv_fusion, False)

    @unittest.skipUnless(bool(
            tf.version.VERSION.find('1.15.0-up') != -1 or tf.version.VERSION >= '2.1.0'), 'not supported the current tf version.')
    @disable_random()
    def test_conv_biasadd_relu6_fusion(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        x_pad = tf.pad(x, paddings, "CONSTANT")
        conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 16, 16],
                                                 initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(x_pad, conv_weights, strides=[1, 2, 2, 1], padding="VALID")
        normed = tf.compat.v1.layers.batch_normalization(conv)

        relu6 = tf.nn.relu6(normed, name='op_to_store')

        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])
            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization('fake_yaml.yaml')
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.eval_dataloader = common.DataLoader(dataset)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = output_graph_def
            output_graph = quantizer()
            found_conv_fusion = True

            for i in output_graph.graph_def.node:
                if i.op == 'Relu6':
                    found_conv_fusion = False
                    break
            self.assertEqual(found_conv_fusion, True)

    @disable_random()
    def test_conv_biasadd_add_relu_fusion(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        top_relu = tf.nn.relu(x)

        conv_weights2 = tf.compat.v1.get_variable("weight2", [3, 3, 16, 16],
                                                  initializer=tf.compat.v1.random_normal_initializer())
        conv2 = tf.nn.conv2d(top_relu, conv_weights2, strides=[1, 2, 2, 1], padding="SAME")
        normed2 = tf.nn.bias_add(conv2, tf.constant([3.0, 1.2,1,2,3,4,5,6,7,8,9,0,12,2,3,4]))
        relu = tf.nn.relu(normed2 + tf.constant([3.0]))
        relu6 = tf.nn.relu6(relu, name='op_to_store')

        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization('fake_yaml.yaml')
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.eval_dataloader = common.DataLoader(dataset)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = output_graph_def
            output_graph = quantizer()

            found_conv_fusion = False

            for i in output_graph.graph_def.node:
                if i.op == 'QuantizedConv2DWithBiasAndReluAndRequantize':
                    found_conv_fusion = True
                    break

            self.assertEqual(found_conv_fusion, True)

    @disable_random()
    def test_conv_biasadd_addv2_relu_fusion(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        top_relu = tf.nn.relu(x)
        paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        x_pad = tf.pad(top_relu, paddings, "CONSTANT")
        conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 16, 16],
                                                 initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(x_pad, conv_weights, strides=[1, 2, 2, 1], padding="VALID")
        normed = tf.compat.v1.layers.batch_normalization(conv)
        # relu = tf.nn.relu(normed)

        conv_weights2 = tf.compat.v1.get_variable("weight2", [3, 3, 16, 16],
                                                  initializer=tf.compat.v1.random_normal_initializer())
        conv2 = tf.nn.conv2d(top_relu, conv_weights2, strides=[1, 2, 2, 1], padding="SAME")
        normed2 = tf.compat.v1.layers.batch_normalization(conv2)
        # relu2 = tf.nn.relu(normed2)
        add = tf.raw_ops.AddV2(x=normed, y=normed2, name='addv2')
        relu = tf.nn.relu(add)
        relu6 = tf.nn.relu6(relu, name='op_to_store')

        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization('fake_yaml.yaml')
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.eval_dataloader = common.DataLoader(dataset)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = output_graph_def
            output_graph = quantizer()

            found_conv_fusion = False

            for i in output_graph.graph_def.node:
                if i.op == 'QuantizedConv2DWithBiasSignedSumAndReluAndRequantize':
                    found_conv_fusion = True
                    break

            self.assertEqual(found_conv_fusion, True)

    @disable_random()
    def test_conv_fusion_with_last_matmul(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        top_relu = tf.nn.relu(x)
        # paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        # x_pad = tf.pad(top_relu, paddings, "CONSTANT")
        conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 16, 16],
                                                 initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(top_relu, conv_weights, strides=[1, 2, 2, 1], padding="VALID")
        normed = tf.compat.v1.layers.batch_normalization(conv)

        relu = tf.nn.relu(normed)
        pooling = tf.nn.max_pool(relu, ksize=1, strides=[1, 2, 2, 1], padding="SAME")
        reshape = tf.reshape(pooling, [-1, 3136])

        y_data = np.random.random([3136, 1])

        y = tf.constant(y_data, dtype=tf.float32, shape=[3136, 1])
        z = tf.matmul(reshape, y)
        relu1 = tf.nn.relu(z)
        y_data_1 = np.random.random([1, 1])
        y_1 = tf.constant(y_data_1, dtype=tf.float32, shape=[1, 1])

        z_2nd_matmul = tf.matmul(relu1, y_1)
        relu6 = tf.nn.relu6(z_2nd_matmul, name='op_to_store')

        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization('fake_yaml.yaml')
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.eval_dataloader = common.DataLoader(dataset)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = output_graph_def
            output_graph = quantizer()

            quantize_v2_count = 0
            for i in output_graph.graph_def.node:
                if i.op == 'QuantizeV2':
                    quantize_v2_count += 1
                    break

            self.assertEqual(quantize_v2_count, 1)

    @disable_random()
    def test_conv_fusion_with_last_conv(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        top_relu = tf.nn.relu(x)
        conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 16, 16],
                                                 initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(top_relu, conv_weights, strides=[1, 2, 2, 1], padding="VALID")
        normed = tf.compat.v1.layers.batch_normalization(conv)

        relu = tf.nn.relu(normed)
        pooling = tf.nn.max_pool(relu, ksize=1, strides=[1, 2, 2, 1], padding="SAME")
        conv_weights_2 = tf.compat.v1.get_variable("weight2", [3, 3, 16, 16],
                                                   initializer=tf.compat.v1.random_normal_initializer())
        conv2 = tf.nn.conv2d(pooling, conv_weights_2, strides=[1, 2, 2, 1], padding="VALID")
        conv_weights_3 = tf.compat.v1.get_variable("weight3", [3, 3, 16, 16],
                                                   initializer=tf.compat.v1.random_normal_initializer())
        relu2 = tf.nn.relu(conv2)
        conv3 = tf.nn.conv2d(relu2, conv_weights_3, strides=[1, 2, 2, 1], padding="VALID")

        relu3 = tf.nn.relu(conv3)
        relu6 = tf.nn.relu6(relu3, name='op_to_store')

        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization('fake_yaml.yaml')
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.eval_dataloader = common.DataLoader(dataset)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = output_graph_def
            output_graph = quantizer()

            quantize_v2_count = 0
            for i in output_graph.graph_def.node:
                if i.op == 'QuantizeV2':
                    quantize_v2_count += 1
                    break

            self.assertEqual(quantize_v2_count, 1)

    @disable_random()
    def test_conv_fusion_with_max_pooling(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")

        relu = tf.nn.relu(x)
        pooling = tf.nn.max_pool(relu, ksize=1, strides=[1, 2, 2, 1], padding="SAME")
        conv_weights = tf.compat.v1.get_variable("weight2", [3, 3, 16, 16],
                                                   initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(pooling, conv_weights, strides=[1, 2, 2, 1], padding="VALID")
        biasadd = tf.compat.v1.layers.batch_normalization(conv, name='op_to_store')
        out_name = biasadd.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            output_graph_def = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization('fake_yaml.yaml')
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.eval_dataloader = common.DataLoader(dataset)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = output_graph_def
            output_graph = quantizer()

            quantized_pool_data_type = None
            quantized_conv_data_type = None
            for i in output_graph.graph_def.node:
                if i.op.find("QuantizedMaxPool") != -1:
                    quantized_pool_data_type = i.attr['T'].type
                if i.op.find("QuantizedConv2D") != -1:
                    quantized_conv_data_type = i.attr['Tinput'].type

            self.assertNotEqual(quantized_pool_data_type, None)
            self.assertEqual(quantized_pool_data_type, quantized_conv_data_type)
class TestGraphConvFusion(unittest.TestCase):
    rn50_fp32_pb_url = 'https://storage.googleapis.com/intel-optimized-tensorflow/models/v1_6/resnet50_fp32_pretrained_model.pb'
    pb_path = '/tmp/.neural_compressor/resnet50_fp32_pretrained_model.pb'
    inputs = ['input']
    outputs = ['predict']

    op_wise_config = {
        "v0/resnet_v13/conv14/conv2d/Conv2D": (False, 'minmax', False, 7.0),
        "v0/resnet_v13/conv11/conv2d/Conv2D": (False, 'minmax', False, 7.0),
        "v0/resnet_v17/conv27/conv2d/Conv2D": (False, 'minmax', False, 7.0)
    }

    @classmethod
    def setUpClass(self):
        if not os.path.exists(self.pb_path):
            os.system('mkdir -p /tmp/.neural_compressor && wget {} -O {} '.format(self.rn50_fp32_pb_url, self.pb_path))
        self.input_graph = tf.compat.v1.GraphDef()
        with open(self.pb_path, "rb") as f:
            self.input_graph.ParseFromString(f.read())

    def test_conv_biasadd_relu_fusion(self):
        tf.compat.v1.disable_eager_execution()

        self._tmp_graph_def = graph_util.remove_training_nodes(self.input_graph, self.outputs)

        self._tmp_graph_def = StripUnusedNodesOptimizer(self._tmp_graph_def,
                                                        self.inputs, self.outputs).do_transformation()

        self._tmp_graph_def = FoldBatchNormNodesOptimizer(self._tmp_graph_def).do_transformation()
        op_wise_sequences = TensorflowQuery(local_config_file=os.path.join(
            os.path.dirname(__file__), "../neural_compressor/adaptor/tensorflow.yaml")).get_eightbit_patterns()

        output_graph, _ = QuantizeGraphForIntel(self._tmp_graph_def, self.outputs,
                                             self.op_wise_config, op_wise_sequences,
                                             'cpu').do_transform()

        node_name_type_mapping = {}
        for i in output_graph.node:
            node_name_type_mapping[i.name] = i.op

        should_disable_sum_node_name = 'v0/resnet_v17/conv27/conv2d/Conv2D_eightbit_quantized_conv'
        should_enable_sum_node_name = 'v0/resnet_v13/conv11/conv2d/Conv2D_eightbit_quantized_conv'
        should_disable_sum_flag = should_disable_sum_node_name in node_name_type_mapping and node_name_type_mapping[
            should_disable_sum_node_name] == 'QuantizedConv2DWithBias'
        should_enable_sum_flag = should_enable_sum_node_name in node_name_type_mapping and node_name_type_mapping[
            should_enable_sum_node_name] == 'QuantizedConv2DWithBiasSumAndRelu'
        self.assertEqual(should_enable_sum_flag, True)
        self.assertEqual(should_disable_sum_flag, True)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration pruning module."""

from typing import Any, Dict, List, Optional

from neural_compressor.ux.utils.json_serializer import JsonSerializer


class Magnitude(JsonSerializer):
    """Configuration Magnitude class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Magnitude class."""
        super().__init__()
        # [Optional]
        self.weights: Optional[List[Any]] = data.get("weights", None)
        # [Optional] One of "per_channel", "per_tensor"
        self.method = data.get("method", None)
        # [Optional] 0 <= s < 1
        self.init_sparsity = data.get("init_sparsity", None)
        # [Optional] # 0 <= s < 1
        self.target_sparsity = data.get("target_sparsity", None)
        # [Optional]
        self.start_epoch = data.get("start_epoch", None)
        # [Optional]
        self.end_epoch = data.get("end_epoch", None)


class Pruning(JsonSerializer):
    """Configuration Pruning class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Pruning class."""
        super().__init__()
        # [Optional]
        self.magnitude = Magnitude(data.get("magnitude", {}))
        # [Optional]
        self.start_epoch = data.get("start_epoch", None)
        # [Optional]
        self.end_epoch = data.get("end_epoch", None)
        # [Optional]
        self.frequency = data.get("frequency", None)
        # [Optional] 0 <= s < 1
        self.init_sparsity = data.get("init_sparsity", None)
        # [Optional] 0 <= s < 1
        self.target_sparsity = data.get("target_sparsity", None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util


@pattern_registry(pattern_type='QKVReshape')
class QKVReshape(Pattern):
    def __call__(self, model):

        pattern_mapping_config = {
            'QKVReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'StridedSlice'), (3, 'Pack'), (4, 'Reshape')],
                               [(0, 'Shape'), (2, 'StridedSlice'), (3, 'Pack')]],
                        'out': [[(0, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 4
                    },
                    'input_tensors': {
                        0: [[{
                            4: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [3]
                },

                # bert_base_mrpc
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'StridedSlice'), (2, 'Pack'), (3, 'Reshape')]],
                        'out': [[(0, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3
                    },
                    'input_tensors': {
                        0: [[{
                            3: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            3: [0]
                        }], [[0], 1]]
                    },
                    'returns': [2]
                },

                # distil_bert_base
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (3, 'Concat'),
                                (4, 'Reshape')]],
                        'out': [[(0, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 4
                    },
                    'input_tensors': {
                        0: [[{
                            4: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [3]
                },

                # geminet
                {
                    'patterns': {
                        'in': [[(0, 'MatMulWithBias'), (1, 'Reshape')]],
                        'out': [[(0, 'MatMulWithBias'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1
                    },
                    'input_tensors': {
                        0: [[
                            {0: [0]}, {0: [1]}, {0: [2]}
                        ], [[0, 1, 2], 3]],
                        1: [[
                            {'input_data': [0]}
                        ], [[1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            1: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1]
                },
            ]
        }

        def _set_attr(head_num, node_names, model):
            attr = OrderedDict()
            attr['dst_shape'] = '-1,-1,' + str(head_num) + ',64'
            attr['dims'] = '0,1'
            
            reshape_node_idx = model.get_node_id(node_names[0])
            model.nodes[reshape_node_idx].attr = attr

        for i in range(len(pattern_mapping_config['QKVReshape'])-1):
            pattern_dict = pattern_mapping_config['QKVReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping(pattern_dict, model)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    pack_node = ret_old_nodes[j][0]
                    head_num = int(pack_node.input_tensors[-2].data)
                    _set_attr(head_num, new_node_names[j], model)
                
                return model
        
        # special reshape node, like has '0,0,12,64' dst_shape attr
        pattern_dict = pattern_mapping_config['QKVReshape'][-1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping(pattern_dict, model)
        if len(new_node_names) != 0:
            for j in range(len(new_node_names)):
                reshape_node = ret_old_nodes[j][1]
                try:
                    dst_shape = reshape_node.attr['dst_shape']
                except:
                    dst_shape = None
                if dst_shape != None and dst_shape.split(',')[0] == '0':
                    head_num = int(dst_shape.split(',')[-2])
                    reshape_node_idx = model.get_node_id(new_node_names[j][1])
                    model.nodes[reshape_node_idx].attr = OrderedDict({
                        'dst_shape': '-1,-1,'+str(head_num)+',64', 'dims': '0,1'})
                mat_node_idx = model.get_node_id(new_node_names[j][0])
                model.nodes[mat_node_idx].attr = ret_old_nodes[j][0].attr
            
            return model

        return model



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""File browser."""

import os
from typing import Any, Dict, List

from neural_compressor.ux.components.model.repository import ModelRepository
from neural_compressor.ux.utils.exceptions import (
    AccessDeniedException,
    ClientErrorException,
    NotFoundException,
)
from neural_compressor.ux.utils.utils import is_dataset_file, is_hidden, verify_file_path


def get_directory_entries(
    data: Dict[str, Any],
) -> Dict[str, Any]:
    """Get directory entries."""
    try:
        path = get_requested_path(data)
        verify_file_path(path)
        contents = get_non_hidden_directory_entries(path)

        contents = filter_requested_entries(contents, get_filter_value(data))

        return {
            "path": path,
            "contents": sort_entries(contents),
        }
    except PermissionError as err:
        raise AccessDeniedException(err)
    except FileNotFoundError as err:
        raise NotFoundException(err)
    except NotADirectoryError as err:
        raise ClientErrorException(err)


def get_requested_path(data: Dict[str, Any]) -> str:
    """Get name of requested filter."""
    try:
        path = data["path"][0]
    except KeyError:
        path = "."

    return os.path.abspath(path)


def get_non_hidden_directory_entries(path: str) -> List:
    """Build a list of entries for path."""
    entries = []

    with os.scandir(path) as it:
        for entry in it:
            if is_hidden(entry.path):
                continue
            if entry.is_dir():
                entries.append(create_dir_entry(entry))
            if entry.is_file():
                entries.append(create_file_entry(entry))

    return entries


def create_dir_entry(entry: os.DirEntry) -> Dict:
    """Build a Directory entry."""
    return create_entry(entry.path, True)


def create_file_entry(entry: os.DirEntry) -> Dict:
    """Build a File entry."""
    return create_entry(entry.path, False)


def create_entry(path: str, is_directory: bool) -> Dict:
    """Build an entry."""
    entry_type = "directory" if is_directory else "file"
    return {
        "name": path,
        "type": entry_type,
        "is_model": ModelRepository.is_model_path(path),
    }


def sort_entries(entries: List) -> List:
    """Sort entries, directories first, files second."""
    entries.sort(key=lambda e: f"{e['type']} {e['name']}")
    return entries


def should_show_files(data: Dict[str, Any]) -> bool:
    """Decide if files be returned."""
    return get_setting_value("files", True, data)


def should_show_only_model_files(data: Dict[str, Any]) -> bool:
    """Decide if files be returned."""
    return get_setting_value("models_only", False, data)


def get_setting_value(setting: str, default: bool, data: Dict[str, Any]) -> bool:
    """Get bool value from parameters."""
    try:
        not_default = repr(not default).lower()
        tested_value = data[setting][0].lower()

        # opposite of default MUST be provided explicit to return not default
        return not default if not_default == tested_value else default
    except KeyError:
        return default


def get_filter_value(data: Dict[str, Any]) -> str:
    """Get name of requested filter."""
    try:
        return data["filter"][0].lower()
    except KeyError:
        return ""


def filter_requested_entries(entries: List, filter_name: str) -> List:
    """Filter list of entries using provided filter."""
    filter_map = {
        "models": is_model_or_directory_entry,
        "datasets": is_dataset_or_directory_entry,
        "directories": is_directory_entry,
    }

    requested_filter = filter_map.get(filter_name)
    if requested_filter is None:
        return entries

    return list(filter(requested_filter, entries))


def is_directory_entry(entry: Dict) -> bool:
    """Return if given entry is for directory."""
    return "directory" == entry["type"]


def is_model_or_directory_entry(entry: Dict) -> bool:
    """Return if given entry should be shown on model list."""
    return entry["is_model"] or is_directory_entry(entry)


def is_dataset_or_directory_entry(entry: Dict) -> bool:
    """Return if given entry should be shown on dataset list."""
    return is_dataset_file(entry["name"]) or is_directory_entry(entry)



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TensorFlow model utils."""

try:
    import tensorflow.compat.v1 as tf_v1
except ImportError:
    import tensorflow as tf_v1

from typing import Any, List


def get_input_shape(graph_def: Any, fix_dynamic_shape: int) -> List[int]:
    """Get input shape of passed graph."""
    graph = tf_v1.Graph()
    with graph.as_default():  # pylint: disable=not-context-manager
        tf_v1.import_graph_def(graph_def, name="")
    for node in graph.as_graph_def().node:  # pylint: disable=no-member
        if node.op == "Placeholder":
            node_dict = {}
            node_dict["type"] = tf_v1.DType(node.attr["dtype"].type).name

            if node_dict["type"] != "bool":
                # convert shape to list
                try:
                    _shape = list(tf_v1.TensorShape(node.attr["shape"].shape))
                    if tf_v1.__version__ >= "2.0.0":
                        node_dict["shape"] = [
                            item if item is not None else fix_dynamic_shape for item in _shape
                        ]
                    else:
                        node_dict["shape"] = [
                            item.value if item.value is not None else fix_dynamic_shape
                            for item in _shape
                        ]
                    # if shape dimension > 1, suppose first dimension is batch-size
                    if len(node_dict["shape"]) > 1:
                        node_dict["shape"] = node_dict["shape"][1:]
                except ValueError:
                    _shape = [fix_dynamic_shape, fix_dynamic_shape, 3]
                    node_dict["shape"] = _shape

            else:  # deal with bool dtype inputs, now assign bool dtype input False value
                node_dict["shape"] = None
                node_dict["value"] = False

    return node_dict["shape"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='OutputData')
class OutputData(Pattern):
    def __call__(self, model):

        # make the output_data node in graph
        last_node = model.nodes[-1]
        input_tensors = copy.deepcopy(last_node.output_tensors)
        output_data_node = util.construct_node('output_data',
                                               'Output',
                                               input_tensors=input_tensors)
        model.insert_nodes(len(model.nodes), [output_data_node])
        model.nodes[-1].attr = None
        
        return model
